#ifndef AHSPRING_H
#define AHSPRING_H
//*****************************************************************************
//* =====================
//*  AHSpring
//* =====================
//*  
//* (Description)
//*    e+e- -> ZA generator
//*
//* (Update Record)
//*    2017/10/31  K.Fujii	Original version.
//*
//*****************************************************************************

#include "TNamed.h"
#include "JSFModule.h"
#include "JSFBases.h"
#include "JSFSpring.h"
#include "JSFBeamGeneration.h"

#include "HELLib.h"
#include "GENLib.h"

//_______________________________________________________________________
// =====================
//  class AHBases
// =====================
//-----------------------------------------------------------------------
class AHBases : public JSFBases {
friend class AHSpringBuf;
public:
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
  AHBases(const char *name  = "AHBases", 
          const char *title = "AH Bases");
  virtual ~AHBases();

  // ----------------------
  //  Getters and Setters
  // ----------------------
  Double_t GetEcmInit  ()           const { return fEcmInit;   }
  Double_t GetMass     ()           const { return fMass;      }
  Double_t GetEta      ()           const { return fEta;       }
  Double_t GetPhi      ()           const { return fPhi;       }
  Double_t GetZBoost   ()           const { return fZBoost;    }
  Double_t GetEcmIP    ()           const { return fEcmIP;     }

  void     SetEcmInit  (Double_t ecm    ) { fEcmInit   = ecm;  }
  void     SetMass     (Double_t m      ) { fMass      = m;    }
  void     SetISR      (Bool_t b = kTRUE) { fISR       = b;    }
  void     SetBeamStr  (Bool_t b = kTRUE) { fBeamStr   = b;    }
  void     SetBeamWidth(Double_t w      ) { fBeamWidth = w;    }
  void     SetPole     (Double_t p      ) { fPole      = p;    }

  // ----------------------
  //   Base class methods
  // ----------------------
  virtual void     Userin();   // Bases user initialization
  virtual void     Userout();  // Bases user output 

  Double_t Func();     // Bases integration function.

  // ----------------------
  //   Utility methods
  // ----------------------
private:
  void     SelectHelicities(Double_t &weight);

  Double_t  DSigmaDX     (GENBranch &cmbranch);
  Double_t  AmpSquared   (GENBranch &cmbranch);
  Complex_t FullAmplitude();
  Complex_t AmpEEtoAH    (const HELFermion &em,
                          const HELFermion &ep,
                          const HELVector  &af,
                          const HELScalar  &hf);

private:
  // --------------------------------------------------------------------
  //  Data Members
  // --------------------------------------------------------------------
  // ----------------
  //  Job parameters
  // ----------------
#if 0
  Double_t fLambda;         // lambda
  Double_t fC0;             // C_0
  Double_t fC1;             // C_1
  Double_t fC2;             // C_2
  Double_t fC3;             // C_3
#endif
  Double_t fMass;           // m_h    : mass  of h
  Double_t fEcmInit;        // Initial Ecm
  Int_t    fISR;            // ISR on?
  Int_t    fBeamStr;        // Beamstrahlung on?
  Double_t fBeamWidth;      // Beam width relative to Ebm(nominal)
  Double_t fPole;           // electron polarization
  Int_t    fNCALL;          // # points in a cell
  Double_t fACC1;           // accuracy for grid optimization
  Double_t fACC2;           // accuracy for integration
  Int_t    fITMX1;          // # iterations for grid optimization
  Int_t    fITMX2;          // # iterations for integration

  // ----------------
  //  Particle Data
  // ----------------
  GENPDTZBoson *fZBosonPtr;       //! PD table entry of "Z"
  GENPDTPhoton *fPhotonPtr;       //! PD table entry of "Gamma"

  // ----------------
  //  Event info
  // ----------------
  Double_t       fZBoost;         // p_z(cm)      in lab frame
  Double_t       fEcmIP;          // Ecm after B-strahlung
  Double_t       fEta;            // eta for photon
  Double_t       fPhi;            // phi_z        in cm  frame
  Double_t       fQ2AH;           // q^2 of AH system
  Int_t          fHelInitial[2];  // initial state helicities
  Int_t          fHelFinal  [2];  // final   state helicities
  Int_t          fJCombI;         // intial helicity combination
  Int_t          fJCombF;         // final  helicity combination
  ANL4DVector    fK[2];           // [0,1] = [e-, e+]
  ANL4DVector    fP[2];           // [0,1] = [a, h ]
  Double_t       fM[2];           // [0,1] = [0, mh]
  Double_t       fSh;             // sin(theta/2) for photon
  Double_t       fCh;             // cos(theta/2) for photon

  // ----------------
  //  Integ. vars
  // ----------------
  Double_t       fHelCombInitial; // initial state helicity combination
  Double_t       fHelCombFinal;   // final   state helicity combination
  Double_t       fXEta;           // eta for photon
  Double_t       fXPhi;           // phi for photon

  Double_t       fR_BW_m;         //! naturl Ebeam spread for e-
  Double_t       fR_BW_p;         //!                     for e+
  Double_t       fR_BS_m;         //! beamstrahlung       for e-
  Double_t       fR_BS_p;         //!                     for e+

  Double_t       fR_ISR_var;      //! initial state radiation
  Double_t       fR_ISR_side;     //! from which side ? (e-/e+) 

  TFile   *fBeamFile;         //! Beamstrahlung data file
  JSFBeamGenerationCain *fBM; //! Beamstrahlung data

  ClassDef(AHBases, 1) // Bases for e+e- -> XX process
};

class AHSpring;

//_______________________________________________________________________
// =====================
//  class AHSpringBuf
// =====================
//-----------------------------------------------------------------------
class AHSpringBuf : public JSFSpringBuf {
public:
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
  AHSpringBuf(const char *name   = "AHSpringBuf", 
              const char *title  = "AH Spring test event buffer",
	      AHSpring   *spring = 0)
	   : JSFSpringBuf(name,title,(JSFSpring*)spring) {} 
  virtual ~AHSpringBuf() {}

  // ----------------------
  //  Getters and Setters
  // ----------------------
  Double_t GetEta      ()           const { return fEta;       }
  Double_t GetPhi      ()           const { return fPhi;       }
  Double_t GetZBoost   ()           const { return fZBoost;    }
  Double_t GetEcmIP    ()           const { return fEcmIP;     }

  // ----------------------
  //   Base class methods
  // ----------------------
  virtual Bool_t SetPartons();

private:
  // --------------------------------------------------------------------
  //  Data Members
  // --------------------------------------------------------------------
  // ----------------
  //  Event info
  // ----------------
  Double_t fEta;            // eta for photon
  Double_t fPhi;            // phi for photon in cm  frame
  Double_t fZBoost;         // p_z(cm)      in lab frame
  Double_t fEcmIP;          // Ecm after B-strahlung

  ClassDef(AHSpringBuf, 1)  // AHSpring event buffer
};

//_______________________________________________________________________
// =====================
//  class AHSpring
// =====================
//-----------------------------------------------------------------------
class AHSpring : public JSFSpring {
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
public:
   AHSpring(const char *name  = "AHSpring", 
            const char *title = "AH Spring test",
            AHBases    *bases = 0);
   virtual ~AHSpring();

   ClassDef(AHSpring, 1)  // AHSpring class
};
#endif
