//*****************************************************************************
//* =====================
//*  EASpring
//* =====================
//*  
//* (Description)
//*    e- gamma --> e- gamma generator
//*
//* (Update Record)
//*    2019/12/25  K.Fujii	Original version.
//*
//*****************************************************************************

#include "JSFSteer.h"
#include "EASpring.h"

#include "TRandom.h"

#include <sstream>
#include <iomanip>
//#define __DEBUG__

//*----------------------------------------------------------------------
//*     Numerical and Natural Constants
//*----------------------------------------------------------------------
#include "GENNumCon.h"

ClassImp(EASpring)
ClassImp(EASpringBuf)
ClassImp(EABases)

//-----------------------------------------------------------------------------
// ==============================
//  class EASpring
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
EASpring::EASpring(const char    *name,
                   const char    *title,
                         EABases *bases)
         : JSFSpring(name, title, bases)
{
  fEventBuf = new EASpringBuf("EASpringBuf",
                              "EASpring event buffer",
                              this);
  if (!bases) { 
    SetBases(new EABases);
  }
}

//_____________________________________________________________________________
// --------------------------
//  D-tor
// --------------------------
EASpring::~EASpring()
{
  delete fEventBuf;
  delete GetBases();
}


//-----------------------------------------------------------------------------
// ==============================
//  class EASpringBuf
// ==============================
//_____________________________________________________________________________
// --------------------------
//  SetPartons
// --------------------------
Bool_t EASpringBuf::SetPartons()
{
  TClonesArray &partons = *fPartons;
  EABases      *bases   = (EABases*)((EASpring*)Module())->GetBases();

  // ----------------------------------------------
  //  Set 4-momenta of final state particles in CM
  // ----------------------------------------------
  fNparton = 2;

  static ANL4DVector *pv = 0;
  if (!pv) {
    pv = new ANL4DVector [fNparton];
  }
  pv[0] = bases->fP[0];
  pv[1] = bases->fP[1];

  // ----------------------------------------------
  //  Boost them to lab. frame
  // ----------------------------------------------

  fZBoost       = bases->GetZBoost();
  fEcmIP        = bases->GetEcmIP();
  fCosTheta     = bases->GetCosTheta();
  fPhi          = bases->GetPhi();
  Double_t elab = TMath::Sqrt(fEcmIP*fEcmIP + fZBoost*fZBoost);
  TVector3 boostv(0.,0.,fZBoost/elab);

  for (Int_t i=0; i<fNparton; i++) pv[i].Boost(boostv);

  // ----------------------------------------------
  //  Set final state parton infomation
  // ----------------------------------------------

  static TVector **qp = 0;
  if (!qp) {
    qp = new TVector* [fNparton];
    for (Int_t i=0; i<fNparton; i++) qp[i] = new TVector(4);
  }
  for (Int_t i=0; i<fNparton; i++) {
    TVector &q = *qp[i];
    q(0) = pv[i].E ();
    q(1) = pv[i].Px();
    q(2) = pv[i].Py();
    q(3) = pv[i].Pz();
  }
  Int_t    ida     = 22;
  Int_t    idf     = bases->fFPtr->GetPID   ();   // PDG code for f
  Double_t mf      = bases->fFPtr->GetMass  ();   // F mass
  Double_t chrg    = bases->fFPtr->GetCharge();   // F charge
  Double_t color   = bases->fFPtr->GetColor ();   // color factor
  Double_t hf      = bases->fHelFinal[0];         // F helicity
  Double_t ha      = bases->fHelFinal[1];         // A helicity
  Int_t    islev   = color > 1. ? 201 : 0;     	  // shower level
  Int_t    icf     = 2;                           // color flux id

  //                                No. PID  Mass  Charge   pv   Nd 1st Mom hel col shower
  new (partons[0]) JSFSpringParton(1, idf,   mf,  chrg, *qp[0],  0, 0,  0,  hf, icf, islev);
  new (partons[1]) JSFSpringParton(2, ida,   0.,    0., *qp[1],  0, 0,  0,  ha, icf, islev);

  return kTRUE ;
}

//-----------------------------------------------------------------------------
// ==============================
//  class EABases
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
EABases::EABases(const char *name, const char *title)
        : JSFBases   (name, title), 
          fPminus    (0.),
          fEminus    (TMath::Sqrt(fPminus*fPminus+kM_e*kM_e)),
          fEgamma    (1.),
          fPole      (0.),
          fPola      (0.),
          fFPtr      (0),
          fZBoost    (0.),
          fEcmIP     (0),
          fCosTheta  (0.),
          fPhi       (0.)
{
  //  Constructor of bases.  Default parameter should be initialized here
  //
  // --------------------------------------------
  //  Get parameters from jsf.conf, if specified
  // --------------------------------------------

  cout << "Init ffbases " << endl;
  
  using namespace std;
  stringstream ins(gJSF->Env()->GetValue("EABases.CosthRange","-1.0 1.0"));
  ins >> fXL[0] >> fXU[0];

  ins.clear();
  ins.str("");
  ins.str(gJSF->Env()->GetValue("EABases.PhiOverPiRange","0.0 2.0"));
  ins >> fXL[1] >> fXU[1];

  ins.clear();
  ins.str("");
  ins.str(gJSF->Env()->GetValue("EABases.Peminus","1."));      // p_e- (1GeV)
  ins >> fPminus;
  fEminus = TMath::Sqrt(fPminus*fPminus+kM_e*kM_e);

  ins.clear();
  ins.str("");
  ins.str(gJSF->Env()->GetValue("EABases.Egamma","1."));       // E_gamma (1GeV)
  ins >> fEgamma;

  ins.clear();
  ins.str("");
  ins.str(gJSF->Env()->GetValue("EABases.Pole","0."));         // electron polarization
  ins >> fPole;

  ins.clear();
  ins.str("");
  ins.str(gJSF->Env()->GetValue("EABases.Pola","0."));         // photon polarization
  ins >> fPola;


  // --------------------------------------------
  //  Registor random numbers
  // --------------------------------------------
  //--
  //  Initial state helicity combination
  //--
  DefineVariable(fHelCombInitial, 0., 1., 0, 1);
  DefineVariable(fHelCombFinal  , 0., 1., 0, 1);
  //--
  //  cos(theta) and phi
  //--
  fXL[1] = fXL[1]*TMath::Pi();
  fXU[1] = fXU[1]*TMath::Pi();

  DefineVariable(fCosTheta , fXL[0], fXU[0], 1, 1);
  DefineVariable(fPhi      , fXL[1], fXU[1], 0, 1);
 
  // --------------------------------------------
  //  Set Bases integration parameters
  // --------------------------------------------
  SetNoOfSample(20000);

  SetTuneValue (1.5);
  SetIteration1(0.01, 5);
  SetIteration2(0.001, 200);

}
// --------------------------
//  D-tor
// --------------------------
EABases::~EABases()
{
  delete fFPtr;
}

//_____________________________________________________________________________
// --------------------------
//  Func
// --------------------------
Double_t EABases::Func()
{
  //  Bases Integrand
  //

  Double_t bsWeight = 1.;      // Jacobian factor
  Double_t eminus   = fEminus; // E_e- after ISR and beamstrahlung
  Double_t egamma   = fEgamma; // E_e+ after ISR and beamstrahlung

  // --------------------------------------------
  //  Beamstrahlung
  // --------------------------------------------
  Double_t mf      = fFPtr->GetMass();
  Double_t pminus  = fPminus;
           fZBoost = pminus - egamma; // P_z of the cm system after ISR and beamstrahlung
  Double_t rs      = TMath::Sqrt((eminus+egamma-fZBoost)*(eminus+pminus));
           fEcmIP  = rs;


  // --------------------------------------------
  //  Select final state
  // --------------------------------------------
  Double_t weight = 1;

  // --------------------------------------------
  //  Check if reduced Ecm enough for prod.
  // --------------------------------------------
  if (fEcmIP < mf) {
    return 0.;
  }

  // --------------------------------------------
  //  Select helicity combination
  // --------------------------------------------
  //  Notice that spin average for e- and a is taken
  //  care of here
  SelectHelicities(weight);
  bsWeight *= weight;

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------
  Double_t sigma = DSigmaDX();

  // --------------------------------------------
  //  Fill plots
  // --------------------------------------------
  ANL4DVector pv[2];
  pv[0] = fP[0];
  pv[1] = fP[1];
  Double_t elab = TMath::Sqrt(fEcmIP*fEcmIP + fZBoost*fZBoost);
  TVector3 boostv(0.,0.,fZBoost/elab);
  for (Int_t i=0; i<2; i++) pv[i].Boost(boostv);

  Double_t eminuslab  = pv[0].E();
  Double_t egammalab  = pv[1].E();
  Double_t csminuslab = pv[0].CosTheta();
  Double_t csgammalab = pv[1].CosTheta();
#ifdef __DEBUG__
  cerr << "Ecm = " << fEcmIP << " Ee-lab = " << eminuslab << " Ea-lab = " << egammalab
	  << " cse-lab = " << csminuslab << " csg-lab = " << csgammalab 
	  << " Esum-lab = " << eminuslab+egammalab << endl;
#endif

  H1Fill( "h01", fEcmIP           , (bsWeight*sigma));
  H1Fill( "h02", fCosTheta        , (bsWeight*sigma));
  H1Fill( "h03", fPhi             , (bsWeight*sigma));
  H1Fill( "h04", (Double_t)fJCombI, (bsWeight*sigma));
  H1Fill( "h05", (Double_t)fJCombF, (bsWeight*sigma));
  H1Fill( "h06", eminuslab        , (bsWeight*sigma));
  H1Fill( "h07", csminuslab       , (bsWeight*sigma));
  H1Fill( "h08", egammalab        , (bsWeight*sigma));
  H1Fill( "h09", csgammalab       , (bsWeight*sigma));
#if 0
  H2Fill( "h10", csminuslab, eminuslab , (bsWeight*sigma));
  H2Fill( "h11", csgammalab, egammalab , (bsWeight*sigma));
  H2Fill( "h12", csminuslab, csgammalab, (bsWeight*sigma));
#endif

  return (bsWeight * sigma);
}

//_____________________________________________________________________________
// --------------------------
//  DSigmaDX
// --------------------------
Double_t EABases::DSigmaDX()
{
  // --------------------------------------------
  //  Phase space
  // --------------------------------------------
  // fEcmIP, fCosTheta, fPhi, fFPtr);
  Double_t rs   = fEcmIP;
  Double_t cosx = fCosTheta;
  Double_t phix = fPhi;
  Double_t mf   = fFPtr->GetMass();
  Double_t mf2  = mf*mf;

  ANL4DVector qcm(rs, 0., 0., 0.);
  GENFrame    cmframe;

  GENPhase2 phaseCM(qcm, mf2, 0., cmframe, cosx, phix, 0);
  Double_t betax = phaseCM.GetBetaBar();
  if (betax <= 0.) return 0.;
  fP[0] = phaseCM.GetFourMomentum(0);
  fP[1] = phaseCM.GetFourMomentum(1);

  Double_t betab  = (1.-kM_e/rs)*(1.+kM_e/rs);
  Double_t pb     = (rs/2)*betab;
  Double_t eb     = TMath::Sqrt(pb*pb+kM_e*kM_e);
  Double_t ea     = pb;
  Double_t beta_e = pb/eb;
  fK[0].SetXYZT(0., 0., pb, eb);
  fK[1].SetXYZT(0., 0.,-pb, ea);

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------
#ifdef __DEBUG__
  cerr << " fP[0] = (" 
       << fP[0].E () << ","
       << fP[0].Px() << ","
       << fP[0].Py() << ","
       << fP[0].Pz() << ")" << endl;
  cerr << " fP[1] = (" 
       << fP[1].E () << ","
       << fP[1].Px() << ","
       << fP[1].Py() << ","
       << fP[1].Pz() << ")" << endl;
  ANL4DVector pcm = fP[0] + fP[1];
  cerr << " pcm = (" 
       << pcm.E () << ","
       << pcm.Px() << ","
       << pcm.Py() << ","
       << pcm.Pz() << ")" << endl;
#endif

  // -------------------
  //  Amplitude squared
  // -------------------
  Double_t amp2 = AmpSquared();

  // -------------------
  //  Put them together
  // -------------------
  static const Int_t    kNbr  = 1;
  static const Double_t kFact = k2Pi/(TMath::Power(k4Pi,3*kNbr));

  Double_t identp = 1.;                            // identical particle factor
  Double_t dPhase = kFact * betax;                 // phase space factor
  Double_t flux   = 1./(4.* eb * ea * (1.+beta_e));// beam flux factor

  Double_t sigma  = identp * flux * amp2 * dPhase; // in [1/GeV^2]
           sigma *= kGeV2fb;                       // now in [fb]

  return sigma;
}

//_____________________________________________________________________________
// --------------------------
//  AmpSquared()
// --------------------------
Double_t EABases::AmpSquared()
{
  Double_t  color = fFPtr->GetColor();
  Complex_t amp   = FullAmplitude();
  Double_t  amp2  = TMath::Power(abs(amp),2) * color;

  return amp2;
}

//_____________________________________________________________________________
// --------------------------
//  FullAmplitude()
// --------------------------
Complex_t EABases::FullAmplitude()
{
   Double_t mf     = fFPtr->GetMass();

   static const Bool_t kIsIncoming = kTRUE;
   static const Bool_t kIsOutgoing = kFALSE;

   HELFermion em(fK[0], kM_e, fHelInitial[0], +1, kIsIncoming);
   HELVector  vi(fK[1],   0., fHelInitial[1], -1);

   HELFermion f (fP[0], mf, fHelFinal [0], +1, kIsOutgoing);
   HELVector  vo(fP[1], 0., fHelFinal [1], +1);

   Complex_t amp = AmpEAtoEA(em, vi, f, vo);

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  AmpEAtoEA()
// --------------------------
Complex_t EABases::AmpEAtoEA(const HELFermion &em,
                             const HELVector  &vi,
			     const HELFermion &f,
			     const HELVector  &vo)
{
   Double_t  qe    = -1.;
   Double_t  ge    = qe*kGe;

   Double_t  qf    = fFPtr->GetCharge();
   Double_t  gf    = qf*kGe;

   //--------------------------------------------------
   // Calculate Amplitudes
   //--------------------------------------------------
   HELFermion eai(em, vi, ge, ge, kM_e, 0.);
   Complex_t ampa = HELVertex(eai, f, vo, gf, gf);

   HELFermion eao(em, vo, ge, ge, kM_e, 0.);
   Complex_t ampat = HELVertex(eao, f, vi, gf, gf);

   Complex_t amp = ampa + ampat;

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  Userin
// --------------------------
void EABases::Userin()
{
  // --------------------------------------------
  //  Initialize Z decay table
  // --------------------------------------------
  // f is e- in this program:
   Int_t ig = 0; // 1st generation
   Int_t i3 = 1; // down type
   Int_t lq = 0; // lepton
   const Char_t   *nam  = kName[lq][i3][ig];
         Int_t     pid  = kPID [lq][i3][ig];
	 Double_t  qf   = lq ? (i3 ? -1./3. : 2./3.) : (i3 ? -1. : 0.);
	 Double_t  spin = 0.5;
	 Double_t  mf   = kM_e; //kMass[lq][i3][ig];
	 Double_t  t3   = 0.5 - i3;
	 Double_t  cf   = lq ? 3. : 1.; 
   if (!fFPtr) fFPtr = new GENPDTEntry(nam, pid, qf, spin, mf, ig+1,  t3, cf);
   fFPtr->DebugPrint();

  // --------------------------------------------
  //  Define some plots
  // --------------------------------------------
  H1Init( "h01", "Ecm"   , 100,     0.,     fEcmIP*1.1 );
  H1Init( "h02", "Costh" , 100, fXL[0],          fXU[0]);
  H1Init( "h03", "Phi"   , 100, fXL[1],          fXU[1]);
  H1Init( "h04", "Helin" ,   4,     0.,              4.);
  H1Init( "h05", "Helot" ,   4,     0.,              4.);
  H1Init( "h06", "Eelab" , 100,     0., fEminus+fEgamma);
  H1Init( "h07", "csalab", 100,   -1.0,            +1.0);
  H1Init( "h08", "Ealab" , 100,     0., fEminus+fEgamma);
  H1Init( "h09", "cselab", 100,   -1.0,            +1.0);
#if 0
  H2Init( "h10", "cselab:eelab" , 50, -1.0, +1.0, 50,  0.0, fEminus+fEgamma);
  H2Init( "h11", "csalab:ealab" , 50, -1.0, +1.0, 50,  0.0, fEminus+fEgamma);
  H2Init( "h12", "csalab:csalab", 50, -1.0, +1.0, 50, -1.0,  +1.0);
#endif
}

//_____________________________________________________________________________
// --------------------------
//  Userout
// --------------------------
void EABases::Userout()
{
  cout << "End of EABases----------------------------------- "  << endl
       << "Peminus              = "  << fPminus  << " [GeV]   "       << endl
       << "Eegamma              = "  << fEgamma  << " [GeV]   "       << endl
       << "Ecm                  = "  << fEcmIP   << " [GeV]   "       << endl
       << "(Pole,Pola)          = (" << fPole << ", " << fPola << ")" << endl
       << "Total Cross section  = "  << GetEstimate()  << " +/- "
                                     << GetError()     << " [fb]"     << endl
       << "Number of iterations = "  << GetNoOfIterate()              << endl;
}

//_____________________________________________________________________________
// --------------------------
//  SelectHelicities
// --------------------------
void EABases::SelectHelicities(Double_t &weight)
{
   static const Int_t kNi = 4;
   static const Int_t kIHelComb[kNi][2] = {{-1, -1},
                                           {-1, +1},
                                           {+1, -1},
                                           {+1, +1}};
   static const Int_t kNf = 4;
   static const Int_t kFHelComb[kNf][2] = {{-1, -1},
                                           {-1, +1},
                                           {+1, -1},
                                           {+1, +1}};
   Double_t helm = (1. - fPole)/2.;
   Double_t hela = (1. - fPola)/2.;
   if (fHelCombInitial < helm || helm == 1.) {
      Double_t helcombi = fHelCombInitial/helm;
      if (helcombi < hela) fJCombI = 0;
      else                 fJCombI = 1;
   } else {
      Double_t helcombi = (fHelCombInitial-helm)/(1.-helm);
      if (helcombi < hela) fJCombI = 2;
      else                 fJCombI = 3;
   }
   fHelInitial[0] = kIHelComb[fJCombI][0];
   fHelInitial[1] = kIHelComb[fJCombI][1];
   fJCombF = (Int_t)(fHelCombFinal*kNf);
   fJCombF = TMath::Min(fJCombF, kNf-1);
   fHelFinal  [0] = kFHelComb[fJCombF][0];
   fHelFinal  [1] = kFHelComb[fJCombF][1];
   weight = kNf;
}
