#ifndef ENTBSPRING_H
#define ENTBSPRING_H
//*****************************************************************************
//* =====================
//*  ENTBSpring
//* =====================
//*  
//* (Description)
//*    e+e- -> entb generator
//*
//* (Update Record)
//*    2018/06/29  K.Fujii	Original version.
//*
//*****************************************************************************

#include "TNamed.h"
#include "JSFModule.h"
#include "JSFBases.h"
#include "JSFSpring.h"
#include "JSFBeamGeneration.h"

#include "HELLib.h"
#include "GENLib.h"

class GENPDTTQuark;

//_______________________________________________________________________
// =====================
//  class ENTBBases
// =====================
//-----------------------------------------------------------------------
class ENTBBases : public JSFBases {
friend class ENTBSpringBuf;
public:
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
  ENTBBases(const char *name  = "ENTBBases", 
            const char *title = "ENTB Bases");
  virtual ~ENTBBases();

  // ----------------------
  //  Getters and Setters
  // ----------------------
  Int_t    GetCP       ()           const { return fCP;        }
  Double_t GetEcmInit  ()           const { return fEcmInit;   }
  Double_t GetTopMass  ()           const { return fTopMass;   }
  Double_t GetXi       ()           const { return fXi;        }
  Double_t GetEta0     ()           const { return fEta0;      }
  Double_t GetEta1     ()           const { return fEta1;      }
  Double_t GetPhi0     ()           const { return fPhi0;      }
  Double_t GetPhi1     ()           const { return fPhi1;      }
  Double_t GetQ2TB     ()           const { return fQ2TB;      }
  Double_t GetCosThetaT()           const { return fCosT;      }
  Double_t GetPhiT     ()           const { return fPhiT;      }
  Double_t GetQ2T      ()           const { return fQ2T;       }
  Double_t GetCosThetaW()           const { return fCosW;      }
  Double_t GetPhiW     ()           const { return fPhiW;      }
  Double_t GetQ2W      ()           const { return fQ2W;       }
  Double_t GetCosThetaF()           const { return fCosF;      }
  Double_t GetPhiF     ()           const { return fPhiF;      }
  Double_t GetZBoost   ()           const { return fZBoost;    }
  Double_t GetEcmIP    ()           const { return fEcmIP;     }

  void     SetEcmInit  (Double_t ecm    ) { fEcmInit   = ecm;  }
  void     SetISR      (Bool_t b = kTRUE) { fISR       = b;    }
  void     SetBeamStr  (Bool_t b = kTRUE) { fBeamStr   = b;    }
  void     SetBeamWidth(Double_t w      ) { fBeamWidth = w;    }
  void     SetPolem    (Double_t p      ) { fPolem     = p;    }
  void     SetPolep    (Double_t p      ) { fPolep     = p;    }

  // ----------------------
  //   Base class methods
  // ----------------------
  virtual void     Userin ();  // Bases user initialization
  virtual void     Userout();  // Bases user output 

  Double_t Func();     // Bases integration function.

  // ----------------------
  //   Utility methods
  // ----------------------
private:
  void     SelectHelicities(Double_t &weight);

  Double_t  DSigmaDX     ();
  Double_t  AmpSquared   ();
  Complex_t FullAmplitude();
  Complex_t AmpEEtoENTB(const HELFermion &em,  // e-
                        const HELFermion &ep,  // e+
                        const HELFermion &emf, // e-
                        const HELFermion &nbf, // nu_e-bar
                        const HELFermion &t,   // t
                        const HELFermion &bb); // b-bar

private:
  // --------------------------------------------------------------------
  //  Data Members
  // --------------------------------------------------------------------
  // ----------------
  //  Job parameters
  // ----------------
  Double_t fEcmInit;        // Initial Ecm
  Double_t fTopMass;        // Initial Ecm
  Int_t    fISR;            // ISR on?
  Int_t    fBeamStr;        // Beamstrahlung on?
  Double_t fBeamWidth;      // Beam width relative to Ebm(nominal)
  Double_t fPolem;          // electron polarization
  Double_t fPolep;          // positron polarization
  Int_t    fWModesLo;       // W decay mode lo;
  Int_t    fWModesHi;       // W decay mode hi;
  Int_t    fNCALL;          // # points in a cell
  Double_t fACC1;           // accuracy for grid optimization
  Double_t fACC2;           // accuracy for integration
  Int_t    fITMX1;          // # iterations for grid optimization
  Int_t    fITMX2;          // # iterations for integration

  // ----------------
  //  Particle Data
  // ----------------
  GENPDTWBoson *fWBosonPtr;       //! PD table entry of "W"
  GENPDTZBoson *fZBosonPtr;       //! PD table entry of "Z"
  GENPDTTQuark *fTopPtr;          //! PD table entry of "t"

  // ----------------
  //  Event info
  // ----------------
  Double_t       fZBoost;         // p_z(cm)      in lab frame
  Double_t       fEcmIP;          // Ecm after B-strahlung
  Double_t       fXi;             // xi
  Double_t       fEta0;           // eta_0 for e- -> e-
  Double_t       fEta1;           // eta_1 for e+ -> nb
  Double_t       fPhi0;           // phi_0 for e- -> e-
  Double_t       fPhi1;           // phi_1 for e+ -> nb
  Double_t       fQ2TB;           // q^2 of the final state tb system
  Double_t       fQ2T;            // q^2 of the final state t
  Double_t       fQ2W;            // q^2 of the final state W
  GENDecayMode  *fWModePtr;       //! pointer to W decay mode
  GENPDTEntry   *f4Ptr;           //! point to up   type W daughter
  GENPDTEntry   *f5Ptr;           //! point to down type W daughter
  Double_t       fCosT;           // cos(theta_t) in the (tb) rest frame
  Double_t       fPhiT;           // phi_t        in the (tb) rest frame
  Double_t       fCosW;           // cos(theta_W) in the top rest frame
  Double_t       fPhiW;           // phi_W        in the top rest frame
  Double_t       fCosF;           // cos(theta_fu) in the W rest frame
  Double_t       fPhiF;           // phi_fu        in the W rest frame
  Int_t          fCP;             // CP flip: (-1,+1) = (nu e+ W-, e- nb W+)
  Int_t          fHelInitial[2];  // initial state helicities
  Int_t          fHelFinal  [6];  // final   state helicities
  Int_t          fJCombI;         // intial helicity combination
  Int_t          fJCombF;         // final  helicity combination
  Int_t          fWMode;          // W decay mode
  ANL4DVector    fK[2];           // [0,1] = [e-, e+]
  ANL4DVector    fP[6];           // [0,1,2,3,4,5] = [e-, nb, bb,  b,  fu, fdb]
  Double_t       fM[6];           // [0,1,2,3,4,5] = [me, mn, mb, mb, mfu, mfd]
  Double_t       fSh0;            // sh0
  Double_t       fCh0;            // ch0

  // ----------------
  //  Integ. vars
  // ----------------
  Double_t       fHelCombInitial; // initial state helicity combination
  Double_t       fHelCombFinal;   // final   state helicity combination
  Double_t       fWDecayMode;     // decay mode selector for W
  Double_t       fXCP;            // CP flip: (nu e+ W-, e- nb W+)
  Double_t       fXXi;            // xi
  Double_t       fXEta0;          // eta_0 for e- -> e-
  Double_t       fXEta1;          // eta_1 for e+ -> nb
  Double_t       fXPhi0;          // phi_0 for e- -> e-
  Double_t       fXPhi10;         // phi_1 - phi_0 for e+ -> nb
  Double_t       fXQ2TB;          // q^2 of the final state (tb) system
  Double_t       fXCosT;          // cos(theta_t) in the (tb) rest frame
  Double_t       fXPhiT;          // phi_t        in the (tb) rest frame
  Double_t       fXQ2T;           // q^2 of the final state top
  Double_t       fXCosW;          // cos(theta_W) in the top rest frame
  Double_t       fXPhiW;          // phi_W        in the top rest frame
  Double_t       fXQ2W;           // q^2 of the final state W
  Double_t       fXCosF;          // cos(theta_fu) in the W rest frame
  Double_t       fXPhiF;          // phi_fu        in the W rest frame

  Double_t       fR_BW_m;         //! naturl Ebeam spread for e-
  Double_t       fR_BW_p;         //!                     for e+
  Double_t       fR_BS_m;         //! beamstrahlung       for e-
  Double_t       fR_BS_p;         //!                     for e+

  Double_t       fR_ISR_var;      //! initial state radiation
  Double_t       fR_ISR_side;     //! from which side ? (e-/e+) 

  Double_t       fXU[6];          //! [0,1,2,3,4,5] = (cost, phit, cosw, phiw, cosf, phif)
  Double_t       fXL[6];          //!

  TFile   *fBeamFile;         //! Beamstrahlung data file
  JSFBeamGenerationCain *fBM; //! Beamstrahlung data

  ClassDef(ENTBBases, 1) // Bases for e+e- -> XX process
};

class ENTBSpring;

//_______________________________________________________________________
// =====================
//  class ENTBSpringBuf
// =====================
//-----------------------------------------------------------------------
class ENTBSpringBuf : public JSFSpringBuf {
public:
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
  ENTBSpringBuf(const char *name   = "ENTBSpringBuf", 
                const char *title  = "ENTB Spring test event buffer",
	        ENTBSpring *spring = 0)
	     : JSFSpringBuf(name,title,(JSFSpring*)spring) {} 
  virtual ~ENTBSpringBuf() {}

  // ----------------------
  //  Getters and Setters
  // ----------------------
  Int_t    GetCP       ()           const { return fCP;        }
  Double_t GetZBoost   ()           const { return fZBoost;    }
  Double_t GetEcmIP    ()           const { return fEcmIP;     }
  Double_t GetXi       ()           const { return fXi;        }
  Double_t GetEta0     ()           const { return fEta0;      }
  Double_t GetEta1     ()           const { return fEta1;      }
  Double_t GetPhi0     ()           const { return fPhi0;      }
  Double_t GetPhi1     ()           const { return fPhi1;      }
  Double_t GetQ2TB     ()           const { return fQ2TB;      }
  Double_t GetCosThetaT()           const { return fCosT;      }
  Double_t GetPhiT     ()           const { return fPhiT;      }
  Double_t GetQ2T      ()           const { return fQ2T;       }
  Double_t GetCosThetaW()           const { return fCosW;      }
  Double_t GetPhiW     ()           const { return fPhiW;      }
  Double_t GetQ2W      ()           const { return fQ2W;       }
  Double_t GetCosThetaF()           const { return fCosF;      }
  Double_t GetPhiF     ()           const { return fPhiF;      }

  // ----------------------
  //   Base class methods
  // ----------------------
  virtual Bool_t SetPartons();

private:
  // --------------------------------------------------------------------
  //  Data Members
  // --------------------------------------------------------------------
  // ----------------
  //  Event info
  // ----------------
  Int_t    fCP;             // CP flip: (-1,+1) = (nu e+ W-, e- nb W+)
  Double_t fZBoost;         // p_z(cm)      in lab frame
  Double_t fEcmIP;          // Ecm after B-strahlung
  Double_t fXi;             // xi
  Double_t fEta0;           // eta_0 for e- -> e-
  Double_t fEta1;           // eta_1 for e+ -> e+
  Double_t fPhi0;           // phi_0 for e- -> e-
  Double_t fPhi1;           // phi_1 - phi_0 for e+ -> e+
  Double_t fQ2TB;           // q^2 of the final state (tb) system
  Double_t fCosT;           // cos(theta_t) in the (tb) rest frame
  Double_t fPhiT;           // phi_t        in the (tb) rest frame
  Double_t fQ2T;            // q^2 of the final state t
  Double_t fCosW;           // cos(theta_W) in the top rest frame
  Double_t fPhiW;           // phi_W        in the top rest frame
  Double_t fQ2W;            // q^2 of the final state W
  Double_t fCosF;           // cos(theta_fu) in the W rest frame
  Double_t fPhiF;           // phi_fu        in the W rest frame

  ClassDef(ENTBSpringBuf, 1)  // ENTBSpring event buffer
};

//_______________________________________________________________________
// =====================
//  class ENTBSpring
// =====================
//-----------------------------------------------------------------------
class ENTBSpring : public JSFSpring {
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
public:
   ENTBSpring(const char *name  = "ENTBSpring", 
              const char *title = "ENTB Spring test",
              ENTBBases  *bases = 0);
   virtual ~ENTBSpring();

   virtual Bool_t Initialize(); // overload the base class method

   ClassDef(ENTBSpring, 1)  // ENTBSpring class
};

//_______________________________________________________________________
// =====================
//  class GENPDTTQuark
// =====================
//-----------------------------------------------------------------------
class GENPDTTQuark: public GENPDTEntry {
 public:

  GENPDTTQuark(Double_t m = 175., GENPDTWBoson *wp = 0);
  virtual ~GENPDTTQuark() {}

 private:
  void     Initialize();
  Double_t GamToFV(Double_t m1, Double_t m2, Double_t vff, Double_t cf);

 private:
  GENPDTWBoson *fWBosonPtr;  // W boson entry

  ClassDef(GENPDTTQuark, 1)  // NR boson class
};


#endif
