//*************************************************************************   
//* ---------------------                      
//* S4CALCluster Class
//* ---------------------   
//* (Description) 
//*     represent calorimeter cluster
//* (Requires)
//*     S4CALHit
//* (Provides)  
//*     S4CALCluster            : made object class
//* (Update Record)
//*     2004/09/03  S.Yamamoto  Original version.
//*     2004/11/15  S.Yamamoto  Add IsOK
//*************************************************************************

#include "S4CALCluster.h"
#include "S4CALExactHit.h"
#include "S4CALPreHit.h"

#include "TVirtualPad.h"
#include "TPolyMarker3D.h"
#include <iostream>

Double_t S4CALCluster::fgHitEmin   = 0.0000001; // Edep cut [GeV]
Double_t S4CALCluster::fgHitTOFmax = 100.;      // TOF  cut [nsec]
Double_t S4CALCluster::fgHitDTmax  = 100.;      // delta t cut [nsec]
Double_t S4CALCluster::fgHitDXmax  = 150.;      // delta x cut [cm]
//Double_t S4CALCluster::fgHitTOFmax = 10.;     // TOF  cut [nsec]
Double_t S4CALCluster::fgEMCut  = 0.9;          // cut for CalibrationFactor
Double_t S4CALCluster::fgCalibFactorForEM = 30. * 1.73 / 1.01 / 1.41; // calibfactor for EM
Double_t S4CALCluster::fgCalibFactorForHD = 30. * 1.30 / 1.07 / 1.06; // calibfactor for HD

Double_t S4CALCluster::fgCalibBERatioForEM = -1.0;
Double_t S4CALCluster::fgCalibBERatioForHD = -1.0;

ClassImp(S4CALCluster)

//=====================================================================
//* constructor -------------------------------------------------------
S4CALCluster::S4CALCluster(S4HYBTrack      *trackp,
                           EClusterType_t   clustertype,
                           EParticleType_t  particletype)
              : TObjArray(50), 
                fCALType(kgNotDecided),  
                fClusterR(0.),
                fClusterTheta(9999.),
                fClusterPhi(9999.),
                fClusterEnergy(0.),
                fClusterMomentum(0.,0.,0.),
                fTracks(1),
                fTime(0.),
                fPosition(0.,0.,0.),
                fFracEM(1.),
                fClusterType(clustertype),
                fParticleType(particletype),
                fDone(kFALSE)
{
   if (trackp) fTracks.Add(trackp);
}

//=====================================================================
//* destructor --------------------------------------------------------

S4CALCluster::~S4CALCluster()
{
}

//=====================================================================
//* Add ---------------------------------------------------------------

void S4CALCluster::Add(const S4CALHit *hitp)
{
   if (fDone) fDone = kFALSE;
   TObjArray::Add(const_cast<S4CALHit *>(hitp));
   fCALType |= (ECALType_t)hitp->GetCALType(); // set CALType
}

//=====================================================================
//* Append ------------------------------------------------------------
Bool_t S4CALCluster::Append(const S4CALHit *hitp)
{
   S4CALExactHit *ehitp  = static_cast<S4CALExactHit *>(
                              const_cast<S4CALHit *>(hitp)->GetExactHitPtr()
                           );
   S4CALExactHit *ehit0p = GetEntries() > 0
                           ? static_cast<S4CALExactHit *>(
                                static_cast<S4CALHit *>(At(0))
                                                ->GetExactHitPtr()
                                )
                           : 0;

   Int_t pHitID     = S4CALHit::GetSortBy() == S4CALHit::kPreHitID
                      ? ehitp->GetPreHitID()       // p-hitid of hit to append
                      : ehitp->GetPreHitPtr() 
                             ->GetPostHitID();
   Int_t currPHitID = ehit0p                       // p-hitid of current hit
                      ? (S4CALHit::GetSortBy() == S4CALHit::kPreHitID
                         ? ehit0p->GetPreHitID()
                         : ehit0p->GetPreHitPtr()
                                 ->GetPostHitID()) // GetEntries > 0
                      : pHitID;                    // GetEntries == 0

   if (currPHitID == pHitID) {            
      Add(hitp);                                   // add this hit to this cluster
      return kTRUE;
   } else {
      return kFALSE;
   }
}

//=====================================================================
//* GetTrackPtr -------------------------------------------------------

S4HYBTrack *S4CALCluster::GetTrackPtr(Int_t i)
{
   return dynamic_cast<S4HYBTrack *>(fTracks[i]);
}

//=====================================================================
//* IsOK --------------------------------------------------------------

Bool_t S4CALCluster::IsOK(S4CALHit &hit)
{
   if (!fDone) {
      if (!(hit.GetEnergyDeposit() > fgHitEmin && hit.GetTOF() < fgHitTOFmax)) hit.Lock();
   } else {
      Double_t dt = hit.GetTOF() - GetTime();
      Double_t dx = (hit.GetHitPosition() - GetPosition()).Mag();
      if (!(hit.GetEnergyDeposit() > fgHitEmin && hit.GetTOF() < fgHitTOFmax && dt < fgHitDTmax && dx < fgHitDXmax)) hit.Lock();
   }
   return !hit.IsLocked();
}

//=====================================================================
//* Update ------------------------------------------------------------

void S4CALCluster::Update()
{
   if (fDone) return;
   TIter nexthit(this);
       
   for (Int_t loop = 0; loop < 2; loop++) {
      Double_t  eem  = 0.;
      Double_t  ecl  = 0.;
      Double_t  tcl  = 1.e20;
      TVector3  xcl(0.,0.,0.);
      S4CALHit *hitp = 0;
      while ((hitp = dynamic_cast<S4CALHit *>(nexthit()))) {
         if (!IsOK(*hitp)) continue;
         Double_t edep = hitp->GetEnergyDeposit();

	 if( fgCalibBERatioForEM > 0.0 && hitp->GetCALType()&kgEcEm ) {
             edep *= fgCalibBERatioForEM; 
         }
	 else if ( fgCalibBERatioForHD > 0.0 && hitp->GetCALType()&&kgEcHd ) {
	     edep *= fgCalibBERatioForHD;
         }

         if (hitp->GetCALType()&kgBaEm || hitp->GetCALType()&kgEcEm) eem += edep;
         ecl += edep;
         xcl += edep * hitp->GetHitPosition(); 

         if (tcl > hitp->GetTOF()) tcl = hitp->GetTOF(); 
      }
      fClusterEnergy = ecl;
      if (fClusterEnergy <= 0.) return;
      fFracEM    = eem/ecl;
      fTime      = tcl;
      fPosition  = xcl;
      fPosition *= 1./fClusterEnergy;

      fDone = kTRUE;
      nexthit.Reset();
   }

   Calibrate();

   fClusterR     = fPosition.Mag(); 
   fClusterTheta = fPosition.Theta();
   fClusterPhi   = fPosition.Phi();

   fClusterMomentum = fClusterEnergy * fPosition.Unit();
}
//=====================================================================
//* Calibrate ---------------------------------------------------------  
                                          
void S4CALCluster::Calibrate()
{ // !!!!! Temporary treatment starts !!!! //

#if 0
   Double_t calibFactor = fCALType&kgBaHd || fCALType&kgEcHd 
                                          ? fgCalibFactorForHD
                                          : fgCalibFactorForEM;
#else
   Double_t calibFactor = fFracEM < fgEMCut ? fgCalibFactorForHD
                                            : fgCalibFactorForEM;
#endif
   fClusterEnergy *= calibFactor;
  // !!!! Temporary trearment ends !!!! //
}

//=====================================================================
//* DebugPrint --------------------------------------------------------  

void S4CALCluster::DebugPrint(const Char_t *opt)
{
#if 0
   S4CALHit      *hp  = static_cast<S4CALHit *>(At(0));
   S4CALExactHit *ehp = static_cast<S4CALExactHit *>(hp->GetExactHitPtr());
   Int_t         pdgc = ehp->GetPDGCode();
#endif
   using namespace std;
   cerr << " -- S4CALCluster::DebugPrint -------------------- " << endl
        << "  tkPtr = " << GetTrackPtr()                        << endl
        << "  clPtr = " << this                                 << endl
        << "  Nhits = " << GetEntries()                         << endl
        << "  Theta = " << GetTheta()                           << endl
        << "  Phi   = " << GetPhi()                             << endl
        << "  R     = " << GetR()                               << endl
        << "  E     = " << GetEnergy()                          << endl
#if 0
        << "  pdgc  = " << pdgc                                 << endl
#endif
        << " ------------------------------------------------ " << endl;
}

//=====================================================================
//* Draw --------------------------------------------------------------  
                                          
void S4CALCluster::Draw(Int_t color, const Char_t *opt)
{
   if (!gPad || !GetEntries()) return;
   gPad->cd();

   TPolyMarker3D *pm3dp = new TPolyMarker3D(GetEntries());
   pm3dp->SetBit(kCanDelete);
   pm3dp->SetMarkerColor(color);

   pm3dp->SetMarkerStyle(20);
   pm3dp->SetMarkerSize(0.1);
   

   Int_t nhits = 0;
   TIter next(this);
   S4CALHit *chp = 0;
   while ((chp = static_cast<S4CALHit *>(next()))) {
      if (chp->IsLocked()) continue;
      TVector3 xv   = chp->GetHitPosition();
      // Double_t edep = chp->GetEnergyDeposit();
      pm3dp->SetPoint(nhits, xv.X(), xv.Y(), xv.Z());
      nhits++;
   }

   pm3dp->Draw();
   gPad->Update();
}


void S4CALCluster::PrintConstants()
{
  std::cout << "=== Constants used by S4CALCluster ===" << std::endl;
  std::cout << "   Edep cut (fgHitEmin)         =" << fgHitEmin << " (GeV)" << std::endl;
  std::cout << "   TOP cut  (fgHitTOFmax)       =" << fgHitTOFmax << " (nsec)" << std::endl;
  std::cout << "   delta t cut (fgHitDTmax)     =" << fgHitDTmax  << " (nsec)" << std::endl;
  std::cout << "   delta x cut (fgHitDXmax)     =" << fgHitDXmax  << " (cm)" << std::endl;
  std::cout << "   cut for Calibration Factor(fgEMCut)="<< fgEMCut << std::endl;
  std::cout << "   Calib. factor for EM(fgCalibFactorForEM)=" << fgCalibFactorForEM << std::endl;
  std::cout << "   Calib. factor for HD(fgCalibFactorForHD)=" << fgCalibFactorForHD << std::endl;
  std::cout << "   Calib. factor Barrel/Endcap ratio for HD(fgCalibBERatioForEM)=" << fgCalibBERatioForEM << std::endl;
  std::cout << "   Calib. factor Barrel/Endcap ratio for HD(fgCalibBERatioForHD)=" << fgCalibBERatioForHD << std::endl;
   
}

