#ifndef __S4CALCLUSTERUTILS__
#define __S4CALCLUSTERUTILS__
//*************************************************************************   
//* ---------------------                      
//* S4CALClusterUtils Class
//* ---------------------   
//* (Description) 
//*     represent calorimeter cluster
//* (Requires)
//*
//* (Provides)  
//*
//* (Update Record)
//*
//*
//*************************************************************************

#include "TVector3.h"
#include "TH1D.h"
#include "TTree.h"
#include "S4CALSmallCluster.h"
#include "S4CALRealCluster.h"
#include "TObject.h"
#include "S4HYBTrack.h"
#include "JSFSteer.h"

#include <utility>
#include <map>
#include <vector>

class S4CALClusterUtils {

 public:
  S4CALClusterUtils();
  virtual ~S4CALClusterUtils();
  
  //=================================================================
  //* Service method
  //-----------------------------------------------------------------
  
 public:
  //---------------------------------------------------------
  // Make sample histogram.
  // Return value has (upper bound, probability).
  // Total number of components is nbins+1 and first
  // component has the pair(lower bound, 0.).
  //---------------------------------------------------------
  static std::map<Double_t,Double_t> MakeSampleHistogram(TTree* sample,
							 Char_t* param,
							 const Char_t* selection = "",
							 //const Char_t* basesel = "",
							 const Int_t nbins = 30,
							 const Double_t max = 30.);

  //---------------------------------------------------------
  // Calculate the probability.
  //---------------------------------------------------------
  static Double_t CalcProbability(const Double_t value,
				  std::map<Double_t,Double_t> histmap);
				  

  //---------------------------------------------------------
  // Calculate the likelihood.
  // L = P1*P2*...*Pn/(P1*P2*...*Pn+B11*B21*...*Bn1+...+B1m*B2m*...*Bnm)
  //   where n = (number of parameters), m = (number of kinds of B.G.)
  //---------------------------------------------------------
  static Double_t CalcLikelihood(const Int_t nparams,     // # of parameters
				 const Int_t nbgkind,     // # of kind of B.G.
				 const TMatrixD sprob,    // signal probability
				 const TMatrixD bgprob);  // bkg B[n][m]


  //---------------------------------------------------------
  // This function merges 2 small clusters at EM/HD boundary.
  // Namely this function finds some pairs which are close
  // each other.
  //---------------------------------------------------------
  static void  EHMatcher(std::map<Int_t, S4CALSmallCluster*> &sclmap,
			 const Double_t matchcut = 10., // distance (cm)
			 const Int_t minhits = 1);
  
  //---------------------------------------------------------
  // Check the distance between mother and child small
  // cluster. If distance is small, return kTRUE.
  //---------------------------------------------------------
  static Bool_t dXdTCut(S4CALSmallCluster *mother,
			S4CALSmallCluster *child,
			const Double_t cutx = 1.e+5,
			const Double_t cutt = 1.e+3);
  
  //---------------------------------------------------------
  // This function checks "whether pos is within a tube or not".
  // The tube will be made using centerpos, axis, tubeR, tubeH
  // (To be exact, we define the tube-like region, it is not
  // realistic tube.)
  //---------------------------------------------------------
  static Bool_t IsWithinTube(const TVector3 pos,
			     const TVector3 centerpos,
			     const TVector3 axis,
			     const Double_t tubeR,
			     const Double_t tubeH);
  
  //-----------------------------------------------------------------
  // Fit the longitudinal shower profile by linear function (a*x + b).
  // The return value is a, error of a, b, error of b, chi2/ndf.
  // Option can be used when you want to check EM/HD region only.
  //-----------------------------------------------------------------
  static std::vector<Double_t> FitByLinear(std::map<Int_t,S4CALSmallCluster*> &sclmap,
					   S4CALSmallCluster* scls,
					   Double_t tubeR = 10.,    // cm
					   Double_t tubeH = 20.,    // cm
					   Char_t   *opt  = "EM");  // use ECAL only
  
  //-----------------------------------------------------------------
  // Fit the longitudinal e-dep profile by quadratic function
  // (a*x^2 + bx + c). The return value is a, error of a, b, error of b,
  // c, error of c, chi2/ndf.
  // Option can be used when you want to check EM/HD region only.
  //-----------------------------------------------------------------
  static std::vector<Double_t> FitByQuadratic(std::map<Int_t,S4CALSmallCluster*> &sclmap,
					      S4CALSmallCluster* scls,
					      Double_t tubeR = 10.,    // cm
					      Double_t tubeH = 20.,    // cm
					      Char_t   *opt  = "EM");  // use ECAL only
  
  //---------------------------------------------------------
  // Calculate the total energy ratio within 2 tubes and
  // return its ratio (Eout/Ein).
  // The tube axis is defined along scl thrust vector and if
  // scl doesn't have a thrust axis then return value is less
  // than 0.
  //---------------------------------------------------------
  static Double_t CalcIsolation(std::map<Int_t,S4CALSmallCluster*> &sclmap,
				S4CALSmallCluster* scl,
				const Double_t tubeRin = 6.,
				const Double_t tubeRout = 30.,
				const Double_t tubeH = 50.);
  
  //---------------------------------------------------------
  // Calculate missing energy (cheated)
  //---------------------------------------------------------
  static void CalcMissingEnergy();
  
  //---------------------------------------------------------
  // Direction checker
  //---------------------------------------------------------
  static Bool_t DirectionIsMatched(TVector3 v1, TVector3 v2, Double_t cutangle = 0.5);
  
  
  //=================================================================
  //* Getter & Setter
  //-----------------------------------------------------------------
  
 public:
  
  static Double_t GetMissingEnergy();
  static Double_t GetNeutrinoEnergy();
  
 public:

  
  
  //=================================================================
  //* Data member
  //-----------------------------------------------------------------
  
 private:
  
  static Double_t fMissingEnergy;       // disappeared energy
  static Double_t fNeutrinoEnergy;      // neutrino energy
  static Bool_t   fDoneCalcMissing;
  static Int_t    fNEvt;                // event number
  
  ClassDef(S4CALClusterUtils, 1)

};
    
#endif
    
