
//*************************************************************************
//* --------------------
//* S4JetMakerBuf
//* --------------------
//* (Description)
//*	This set of classes makes Jets.
//* (Requires)
//*     JSFEventBuf
//* (Provides)
//*     S4JetMakerBuf           : buffering class
//* (Update Record)
//*	2004/10/05  S.Yamamoto	Original version.
//*************************************************************************

#include <iostream>
#include "S4PFOMakerBuf.h"
#include "S4JetMakerBuf.h"
#include "S4JetSet.h"

ClassImp(S4JetMakerBuf)

//=====================================================================
//* constructor -------------------------------------------------------

S4JetMakerBuf::S4JetMakerBuf(S4JetMaker *module, const Char_t *name, 
			     const Char_t *title, Int_t bufsize)
		    : JSFEventBuf(name, title, (JSFModule*)module),
		      fDone(kFALSE),
                      fInputDataBufPtr(0),
                      fOutputBufPtr(0),
                      fBufSize(bufsize),
                      fDefaultYcut(1.e-3),
                      fMaxNjets(0)  
{
}

//=====================================================================
//* destructor --------------------------------------------------------

S4JetMakerBuf::~S4JetMakerBuf()
{
   if (fOutputBufPtr) fOutputBufPtr->Delete();
   delete fOutputBufPtr;
}

//=====================================================================       
//* GetJetSetPtr ------------------------------------------------------      

S4JetSet * S4JetMakerBuf::GetJetSetPtr(Int_t n)
{
   if (!fDone) MakeJets();
   if (!n) n = fMaxNjets; 
   if (n > fMaxNjets) return 0; // no jetset for n jets
   else if (fOutputBufPtr->At(n)) return static_cast<S4JetSet *>(fOutputBufPtr->At(n));

   for (Int_t njets = n+1; njets <= fMaxNjets; njets++) {
      if (fOutputBufPtr->At(njets)) {
         const S4JetSet &js  = *static_cast<S4JetSet *>(fOutputBufPtr->At(njets));
         ANLJetFinder   *jfp =  new ANLJadeEJetFinder(static_cast<const ANLJadeEJetFinder &>(js.GetJetFinder()));
         jfp->ForceNJets(n); 
         StoreJetSet(jfp);
         return static_cast<S4JetSet *>(fOutputBufPtr->At(n)); // event by event
      } 
   }
   return 0;
}


//=====================================================================
//* SetInputDataBuf ---------------------------------------------------

void S4JetMakerBuf::SetInputDataBufPtr(JSFEventBuf *buf)
{
   fDone = kFALSE;
   fInputDataBufPtr = buf;
   if (!fOutputBufPtr) fOutputBufPtr = new TObjArray(fBufSize);
   if (fOutputBufPtr->GetEntries() > 0) fOutputBufPtr->Delete();
   fMaxNjets = 0;
}
//=====================================================================
//* DebugPrint --------------------------------------------------------
 
void S4JetMakerBuf::DebugPrint(const Char_t *)
{
   if (!fOutputBufPtr) return;
   if (!fDone) MakeJets();
  
   if (S4JetSet *jsp = static_cast<S4JetSet *>(fOutputBufPtr->At(fMaxNjets)))
      jsp->DebugPrint();
}

//=====================================================================
//* MakeJets ----------------------------------------------------------

void S4JetMakerBuf::MakeJets()
{
   if (fDone) return; // check if jets already made 

   fDone = kTRUE;
   fOutputBufPtr->SetOwner();        // buffer owns Jets 

   S4PFOMakerBuf &pfomb = *static_cast<S4PFOMakerBuf *>(fInputDataBufPtr);
   if (!pfomb.GetEntries()) return;
   pfomb.UpdatePFOs();
   
   TObjArray     &pfos = *pfomb.GetOutputBufPtr();

   ANLJetFinder *jfp  = new ANLJadeEJetFinder(fDefaultYcut);
   jfp->Initialize(pfos);
   jfp->FindJets();

   StoreJetSet(jfp);
}
//=====================================================================       
//* StoreJetSet -------------------------------------------------------       

void S4JetMakerBuf::StoreJetSet(ANLJetFinder *jfp)
{
   Int_t njets = jfp->GetNjets();
   if (njets > fMaxNjets) fMaxNjets = njets;

   static const Bool_t kToDeleteJF = kTRUE;
   S4JetSet *jsp = new S4JetSet(jfp, kToDeleteJF);
   fOutputBufPtr->AddAt(jsp,njets); // Add jetset to njets-th entry
} 
