
//*************************************************************************   
//* ---------------------                      
//* S4PFObject Class
//* ---------------------   
//* (Description) 
//*     represent Particle Flow Object.
//* (Requires)
//*     S4CALCluster, S4HYBTrack
//* (Provides)  
//*     S4PFObject            : made object class
//* (Update Record)
//*     2004/09/04  S.Yamamoto  Original version.
//*************************************************************************

#include "S4PFObject.h"
#include "S4PFOMakerBuf.h"
#include "S4V0Finder.h"

#include "S4CALExactHit.h"
#include "S4CALPreHit.h"

#include "S4GENMCTruthBuf.h"
#include "S4HYBTrackMaker.h"
#include "S4HISTMCTruthBuf.h"

#include "JSFSteer.h"

#include "TVirtualPad.h"
#include "TPolyMarker3D.h"

#include <iostream>
#if __GNUC__ >= 5 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3)
#include <cstdlib>
#endif

class J4TPCPostHit;

//#define __DEBUG__
//#define __BPLOCK__
//
// *** VERY TEMPORARY **** 
const Double_t S4PFObject::fkgMasses[] = { 0.,        // undecided
				           0.511e-3,  // electrom
                                           0.105658,  // muom
     				           0.,        // photon
				           0.13957,   // pi
				           0.493677,  // K+/-
				           0.49767,   // kl
                                           0.93827,   // proton
					   0.93957};  // neutron
// *** END OF VERY TEMPORARY **** 

Bool_t S4PFObject::fgIsTKExact    = kFALSE;
Bool_t S4PFObject::fgIsEMExact    = kFALSE;
Bool_t S4PFObject::fgIsHDExact    = kFALSE;
Bool_t S4PFObject::fgUsePrimaryBP = kFALSE;

ClassImp(S4PFObject)

//=====================================================================
//* constructor -------------------------------------------------------

S4PFObject::S4PFObject(S4CALCluster *clusterPtr)
           : ANL4DVector(),  
             fClusterPtr(clusterPtr),
             fTrackPtr((S4HYBTrack *)clusterPtr->GetTracks()[0]),
             fPostHitPtr(0),
	     fBreakPointPtr(0),
	     fDaughterPFOsPtr(0),
             fPID(0),
             fPdgcode(0), 
             fXstart(0.,0.,0.),
             fTstart(0.),
             fMass(0.), 
	     fQuality(0.),
             fDone(kFALSE)
{
   if (clusterPtr->IsLocked()) Lock(); // lock itself if cluster is locked
   S4VExactHit *ehtp   = GetExactHitPtr();
   Int_t j4tkid = ehtp->GetTrackID();
   fBreakPointPtr = S4BreakPoint::GetBreakPointPtr(j4tkid);
   if (fBreakPointPtr) fBreakPointPtr->SetPFOPtr(this); 
}

S4PFObject::S4PFObject(S4HYBTrack     *trackPtr)
           : ANL4DVector(),
             fClusterPtr(0),
             fTrackPtr(trackPtr),
             fPostHitPtr(0),
	     fBreakPointPtr(0),
	     fDaughterPFOsPtr(0),
             fPID(0),
             fPdgcode(0), 
             fXstart(0.,0.,0.),
             fTstart(0.),
             fMass(0.),
	     fQuality(0.),
             fDone(kFALSE)
{
  Int_t j4tkid = fTrackPtr->GetTrackID();
  fBreakPointPtr = S4BreakPoint::GetBreakPointPtr(j4tkid);
  if (fBreakPointPtr) fBreakPointPtr->SetPFOPtr(this); 
}

S4PFObject::S4PFObject(S4BreakPoint *breakpointPtr)
           : ANL4DVector(),
	     fClusterPtr(0),
	     fTrackPtr(0),
	     fPostHitPtr(0),
	     fBreakPointPtr(breakpointPtr),
	     fDaughterPFOsPtr(0),
	     fPID(0),
	     fPdgcode(0), 
	     fXstart(0.,0.,0.),
	     fTstart(0.),
	     fMass(0.),
	     fQuality(0.),
	     fDone(kFALSE)
{
}

S4PFObject::S4PFObject(TObjArray *childrenp)
           : ANL4DVector(),
	     fClusterPtr(0),
	     fTrackPtr(0),
	     fPostHitPtr(0),
	     fBreakPointPtr(0),
	     fDaughterPFOsPtr(0),
	     fPID(0),
	     fPdgcode(0),
	     fXstart(0.,0.,0.),
	     fTstart(0.),
	     fMass(0.),
	     fQuality(0.),
	     fDone(kFALSE)
{
  fDaughterPFOsPtr = new TObjArray(*childrenp);
  S4PFObject *childp = static_cast<S4PFObject *>(fDaughterPFOsPtr->At(0));
  if (childp) {
    S4BreakPoint * childbpp = childp->fBreakPointPtr;
    Int_t j4tkid = childbpp ? childbpp->GetParentID() :0;
    fBreakPointPtr = S4BreakPoint::GetBreakPointPtr(j4tkid);
  }
}

//=====================================================================
//* destructor --------------------------------------------------------

S4PFObject::~S4PFObject()
{
  if (fDaughterPFOsPtr) delete fDaughterPFOsPtr; 
}

//=====================================================================       
//* GetCharge ---------------------------------------------------------

Double_t S4PFObject::GetCharge() 
{
  if (fTrackPtr) return fTrackPtr->GetCharge();
  else           return 0;
}
 
//=====================================================================
//* GetProbability ----------------------------------------------------

Double_t S4PFObject::GetProbability(EPID_t pid)
{
  if (!fDone) Update();
  return fProbs[pid]; 
}

//=====================================================================      
//* GetPID ------------------------------------------------------------        

Int_t S4PFObject::GetPID()
{
  if (!fDone) Update();
  return fPID;
}

//=====================================================================
//* IsPid -------------------------------------------------------------

Bool_t S4PFObject::IsPID(EPID_t pid, Double_t prob)
{
   if (!fDone) Update();
   if (fProbs[pid] > prob) return kTRUE;
   else                    return kFALSE;
}

//=====================================================================       
//* GetPDGCode --------------------------------------------------------        
Int_t S4PFObject::GetPDGCode()
{
  if (!fDone) Update();
  return fPdgcode;
}
//=====================================================================
//* GetPosition -------------------------------------------------------

TVector3 S4PFObject::GetPosition(Int_t ip)
{
   if (!fDone) Update();
   
   using namespace std;

   switch (ip) {
   case 0:
      return fXstart;
   case 1:
   default:
      cerr << "S4PFObject::GetPosition >>>>>>>>>>>>>>" << endl
	   << " ip = " << ip << " is not supported! "  << endl
           << " abort "                                 << endl;
      abort();
   } 
}

//=====================================================================
//* SetPosition -------------------------------------------------------

void S4PFObject::SetPosition(const TVector3 &xv, Int_t ip)
{
   using namespace std;

   switch (ip) {
   case 0:
      fXstart = xv;
      break;
   case 1:
   default:
      cerr << "S4PFObject::SetPosition >>>>>>>>>>>>>>" << endl
	   << " ip = " << ip << " is not supported! "  << endl
           << " abort "                                 << endl;
      abort();
   } 
}

//=====================================================================
//* IsExact -----------------------------------------------------------

Bool_t S4PFObject::IsExact(const Char_t *type)
{
   TString st(type);
   st.ToLower();
   switch (st.Data()[0]) {
      case 't': return fgIsTKExact;
      case 'e': return fgIsEMExact;
      case 'h': return fgIsHDExact;
      default:  return kFALSE;
   }
}

//=====================================================================
//* SetIsExact --------------------------------------------------------

void S4PFObject::SetIsExact(const Char_t *type, Bool_t set)
{
   TString st(type);
   st.ToLower();
   switch (st.Data()[0]) {
      case 't': fgIsTKExact = set; break;
      case 'e': fgIsEMExact = set; break;
      case 'h': fgIsHDExact = set; break;
   }
   cerr << ">>>> S4PFObject::SetIsExact(" << type << ")" 
	<< " = " << set << endl;
}

//=====================================================================
//* GetUsePrimaryBP ---------------------------------------------------

Bool_t S4PFObject::GetUsePrimaryBP()
{
   return fgUsePrimaryBP;
}


//=====================================================================
//* SetUsePrimaryBP ---------------------------------------------------

void S4PFObject::SetUsePrimaryBP(Bool_t set)
{
   fgUsePrimaryBP = set;
   cerr << ">>>> S4PFObject::SetUsePrimaryBP = " << set << endl;
}

//=====================================================================       
//* GetTime -----------------------------------------------------------   

Double_t S4PFObject::GetTime(Int_t ip)
{
  if (!fDone) Update();

  using namespace std;

  switch (ip) {
  case 0:
    return fTstart;
  case 1:
  default:
    cerr << "S4PFObject::GetTime >>>>>>>>>>>>>>>>>>" << endl
	 << " ip = " << ip << " is not supported! "  << endl
	 << " abort "                                << endl;
    abort();
  }
}
//=====================================================================
//* GetConeEnergy -----------------------------------------------------

Double_t S4PFObject::GetConeEnergy(const Double_t cth,
                                   const TObjArray *tracks) 
{
  if (!fDone) Update();

  Double_t e = 0.;
  TIter next(tracks);
  S4PFObject *pfop;
  while ((pfop = static_cast<S4PFObject *>(next()))) {
    if (pfop == this || pfop->IsLocked()) continue;
    if (CosTheta(*pfop) > cth) e += pfop->E();
  }
  return e;
}

//=====================================================================
//* GetExactHiPtr -----------------------------------------------------

S4VExactHit* S4PFObject::GetExactHitPtr()
{
  if (!fClusterPtr) return 0;
  S4VExactHit  *ehp  =  static_cast<S4CALHit *>(fClusterPtr->At(0))
    ->GetExactHitPtr() ;
  S4CALPreHit   *prep = static_cast<S4CALExactHit *>(ehp)->GetPreHitPtr();
  S4TPCPostHit  *posp = prep ? prep->GetPostHitPtr() : 0;
  S4VExactHit   *php  = static_cast<S4VExactHit *>(posp)
                        ? static_cast<S4VExactHit *>(posp)
                        : static_cast<S4VExactHit *>(prep);
  return php;
} 

//=====================================================================
//* GetPrimaryBreakPointPtr -------------------------------------------

S4BreakPoint* S4PFObject::GetPrimaryBreakPointPtr()
{
  Int_t j4tkid = fBreakPointPtr ? fBreakPointPtr->GetTrackID() : 0;
  S4BreakPoint *bpp = 0;
  while ((bpp = S4BreakPoint::GetBreakPointPtr(j4tkid))) {
    j4tkid = bpp->GetParentID();
#if 1
    if (bpp->TestBit(BIT(20))) return bpp;
#endif
    if (!j4tkid) return bpp;
  }
  return bpp;
}

//=====================================================================
//* GetPrimaryParticleBreakPoint --------------------------------------
S4BreakPoint* S4PFObject::GetPrimaryParticleBreakPoint()
{
  return GetPrimaryBreakPointPtr();
}

//=====================================================================
//* UpdateWithPrimaryBreakPoint ---------------------------------------

void S4PFObject::UpdateWithPrimaryBreakPoint(TVector3 &p, Double_t &e)
{
  if (IsLocked()) return;
  S4BreakPoint *bpp = GetPrimaryBreakPointPtr();
  if (!bpp || bpp->IsLocked()) {
      Lock();
  } else {
    p       = bpp->GetFourMomentum().Vect();
    fTstart = bpp->GetGlobalTime();
    fXstart = bpp->GetPosition();
    e       = bpp->GetFourMomentum().E();
#if 0
    bpp->LockAllDescendants(); // this may lock this pfo
#endif
    bpp->Lock(); // avoid double counting when bpp has daughter PFOs
  }
}

//=====================================================================
//* UpdateWithBreakPoint ----------------------------------------------

void S4PFObject::UpdateWithBreakPoint(TVector3 &p, Double_t &e) 
{
  if (fgUsePrimaryBP) {
    UpdateWithPrimaryBreakPoint(p, e);
  } else {
    S4BreakPoint *bpp  = fBreakPointPtr;
    if (!bpp || bpp->IsLocked()) {
      Lock();
    } else {
      p       = bpp->GetFourMomentum().Vect();
      fTstart = bpp->GetGlobalTime();
      fXstart = bpp->GetPosition();
      e       = bpp->GetFourMomentum().E();
      Int_t         apid = TMath::Abs(bpp->GetParticleID());
      S4BreakPoint *pbpp = GetPrimaryBreakPointPtr();
      if (bpp != pbpp && apid == 2112) e -= bpp->GetFourMomentum().Mag();
      if (fDaughterPFOsPtr) {
         TIter next(fDaughterPFOsPtr);
         S4PFObject *cpfop = 0;
	 while ((cpfop = static_cast<S4PFObject *>(next()))) {
           S4BreakPoint *cbpp = cpfop->GetBreakPointPtr();
           if (cbpp) cpfop->SetPosition(cbpp->GetPosition());
         }
      }
#ifdef __BPLOCK__
      bpp->LockAllDescendants();
      bpp->Lock();
#endif
    }
  }
}

//=====================================================================
//* UpdateV0 ----------------------------------------------------------

void S4PFObject::UpdateV0(TVector3 &p, Double_t &e, Double_t &quality) 
{
  S4PFOMakerBuf *pfombp = (S4PFOMakerBuf *)gJSF->FindModule("S4PFOMaker")->EventBuf();
  TVector3 xv;
  quality = pfombp->GetV0FinderPtr()->FitV0(*this, xv, p, e);
}

//=====================================================================
//* Update ------------------------------------------------------------

void S4PFObject::Update()
{
   if (fDone) return;
   fDone = kTRUE;

   DoPID();
   Int_t pdg  = GetPDGCode();
   Int_t apdg = TMath::Abs(pdg);
   TVector3 p;
   Double_t e = 0;
   Double_t m  = GetMass();

// === Track Update ===
   if (fTrackPtr) {
     if (fgIsTKExact) {
       UpdateWithBreakPoint(p,e);
     } else {
       p       = fTrackPtr->GetMomentum();
       fTstart = fTrackPtr->GetTime();
       fXstart = fTrackPtr->GetPosition();
       e  = TMath::Sqrt(m*m + p*p);
     }
   } 
// === Cluster Update ===
   else if (fClusterPtr) {
     if ((fgIsEMExact && (apdg == 22 || apdg == 11))
         || (fgIsHDExact && (apdg != 22 && apdg != 11))) {
       UpdateWithBreakPoint(p,e);
     } else {
       p  = fClusterPtr->GetMomentum();
       fTstart = fClusterPtr->GetTime();
       e  = p.Mag();
     }
   } 
// === V0 Update ===
   else if (fDaughterPFOsPtr) {
     if (fgIsTKExact) {
       UpdateWithBreakPoint(p,e); 
     } else {
       UpdateV0(p,e,fQuality);
     }
   }
// === Neutrino or Acceptance Hole Update ===  
   else if (fBreakPointPtr) {
     UpdateWithBreakPoint(p,e); 
   }
   else  {
     using namespace std;
     cerr<<"S4PFOclP= "<<fClusterPtr<<endl;
     cerr<<"S4PFOtkP= "<<fTrackPtr<<endl;
     cerr<<"S4PFOposP= "<<fPostHitPtr<<endl;
     cerr<<"S4PFObP= "<<fBreakPointPtr<<endl;
     cerr << "S4PFObject::Update>>>>>>>>>>>>>>>>>>" << endl
	  << " fClusterPtr = 0 && fTraclPtr = 0 "   << endl
	  << " abort "                              << endl;
     abort();
   }

   E()  = e;
   Px() = p.X();
   Py() = p.Y();
   Pz() = p.Z();
}

//=====================================================================
//* DoPID -------------------------------------------------------------  

void S4PFObject::DoPID()
{
   // ***** VERY TEMPORARY ***** //
   for (Int_t i = 0; i < fkgNpid; i++) fProbs[i] = 0.;
#if 0
   if (fClusterPtr) {
      EParticleType_t type = fClusterPtr->GetParticleType();
      if (type&kgElectron_p)  fProbs[PID::kgElectron_id] = 1.;
      if (type&kgPhoton_p)    fProbs[PID::kgPhoton_id]   = 1.;
      if (type&kgHdCharged_p) fProbs[PID::kgPi_id]       = 1.;
      if (type&kgHdNeutral_p) fProbs[PID::kgKL_id]       = 1.;
   } else if (fTrackPtr) {
      fProbs[PID::kgPi_id] = 1.;
   } else {
      using namespace std;
      cerr << "S4PFObject::Update>>>>>>>>>>>>>>>>>>" << endl
	   << " fClusterPtr = 0 && fTraclPtr = 0 "   << endl
	   << " abort "                              << endl;
      // abort();
   }
#else
   // >>>>>>>>>>>>>>>>>>>>>>>> TEMPORARY TREATMENT >>>>>>>>>>>>>>>>>>>>>>>>>>>
   Int_t pdgcode = 0;
   // === Cluster PDG ===

   if (fClusterPtr) {
      S4VExactHit   *php  = GetExactHitPtr();

      pdgcode = php->GetPDGCode();
   } 
   // === Track PDG ===
   else if (fTrackPtr) { 
     if (fgIsTKExact) {
       pdgcode = fBreakPointPtr ? fBreakPointPtr->GetParticleID() : fTrackPtr->GetPDGCode();
     } else {
       pdgcode = fTrackPtr->GetPDGCode();
     }
   }
   // === BreakPoint PDG ===
   else if (fBreakPointPtr) { 
       pdgcode = fBreakPointPtr ? fBreakPointPtr->GetParticleID() : 0;
   }
   else {
      using namespace std;
      cerr << "S4PFObject::Update>>>>>>>>>>>>>>>>>>" << endl
	   << " fClusterPtr = 0 && fTraclPtr = 0 "   << endl
	   << " abort "                              << endl;
      abort();
   }

   fPdgcode = pdgcode;
   pdgcode = TMath::Abs(pdgcode);
   if (pdgcode ==   11) fProbs[PID::kgElectron_id] = 1.;
   if (pdgcode ==   13) fProbs[PID::kgMuon_id]     = 1.;
   if (pdgcode ==   22) fProbs[PID::kgPhoton_id]   = 1.;
   if (pdgcode ==  211) fProbs[PID::kgPi_id]       = 1.;
   if (pdgcode ==  130) fProbs[PID::kgKL_id]       = 1.;
   if (pdgcode ==  321) fProbs[PID::kgK_id]        = 1.;
   if (pdgcode == 2212) fProbs[PID::kgProton_id]   = 1.;
   if (pdgcode == 2112) fProbs[PID::kgNeutron_id]  = 1.;
   // >>>>>>>>>>>>>>>>>>>>END OF TEMPORARY TREATMENT >>>>>>>>>>>>>>>>>>>>>>>>>
#endif
   Double_t sum  = 0.;
   Double_t prob = 0.;
   for (Int_t i = 0; i < fkgNpid; i++) {
      sum += fProbs[i];
      if (fProbs[i] > prob) {
         prob = fProbs[i];
         fPID = (EPID_t)i;
      }
   }
   for (Int_t i = 0; i < fkgNpid; i++) fProbs[i] /= sum;
   fMass = fkgMasses[fPID];  
}

//=====================================================================
//* GetMass -----------------------------------------------------------  

Double_t S4PFObject::GetMass()
{
  if (!fDone) Update();
  return fMass;
}

//=====================================================================
//* DebugPrint --------------------------------------------------------  
                                          
void S4PFObject::DebugPrint(const Char_t *opt)
{ 
   if (!fDone) Update();

   // >>>>>>>>>>>>>>>>>>>>>>>> TEMPORARY TREATMENT >>>>>>>>>>>>>>>>>>>>>>>>>>>
   S4VExactHit  *php  = GetExactHitPtr();
   Int_t    pdg   = 0;
   Int_t    tkid  = 0;
   Int_t    tkidc = 0;
   Double_t Eph   = 0.;
   Double_t Mph   = 0.;
   TVector3 Xpr;
   if (fTrackPtr || php) {
     pdg    = fTrackPtr ? fTrackPtr->GetPDGCode() 
                        : php->GetPDGCode();
     tkid   = fTrackPtr ? fTrackPtr->GetTrackID()
                        : php->GetTrackID();
     tkidc  = php ? php->GetTrackID() : -1;
     Eph    = php ? php->GetTotalEnergy() : 0.;
     Mph    = php ? php->GetMomentum().Mag() : 0.;
     Xpr    = php ? php->GetPosition() : TVector3(0.,0.,0.);
   } else  {
     cerr << "fBreakPointPtr1= " << fBreakPointPtr << endl;
     pdg    = fBreakPointPtr ? fBreakPointPtr->GetParticleID() : 0;
     tkid   = fBreakPointPtr ? fBreakPointPtr->GetTrackID() : 0;
     tkidc  = fBreakPointPtr ? fBreakPointPtr->GetTrackID() : -2;
     Eph    = fBreakPointPtr ? fBreakPointPtr->GetFourMomentum().T() : 0.;
     Mph    = fBreakPointPtr ? fBreakPointPtr->GetFourMomentum().Mag() : 0.;
     Xpr    = fBreakPointPtr ? fBreakPointPtr->GetPosition() : TVector3(0.,0.,0.); 
   }
   S4PFObject *bpfop = 0;
   if (fBreakPointPtr) bpfop = static_cast<S4PFObject *>(fBreakPointPtr->GetPFOPtr());
   Bool_t Block = fBreakPointPtr ? fBreakPointPtr->IsLocked() : 0;
   Bool_t Plock = bpfop ? bpfop->IsLocked() : 0;
   // >>>>>>>>>>>>>>>>>>>>END OF TEMPORARY TREATMENT >>>>>>>>>>>>>>>>>>>>>>>>>
   
   Int_t nchits = fClusterPtr ? fClusterPtr->GetEntries() : 0;
   
   using namespace std;
   cerr << "------------ S4PFObject::DebugPrint --------------" << endl
        << "   this  = "  << this                                << endl
        << "   tPtr  = "  << (void *)fTrackPtr                   << endl
        << "   cPtr  = "  << (void *)fClusterPtr                 << endl
        << "   bPtr  = "  << (void *)fBreakPointPtr              << endl
        << "   dPtr  = "  << (void *)fDaughterPFOsPtr            << endl
        << "   T0    = "  << GetTime()         << "  [nsec]"     << endl
        << "   Xv    = (" << GetPosition().X() << ", "
                          << GetPosition().Y() << ", "
                          << GetPosition().Z() << ") [cm]"       << endl
        << "   E     = "  << E()               << "  [GeV]"      << endl
        << "   Pv    = (" << Px()              << ", "
                          << Py()              << ", " 
                          << Pz()              << ") [GeV]"      << endl
        << "   Q     = "  << GetCharge()                         << endl
        << "   PDG   = "  << pdg                                 << endl
        << "   Pid   = "  << GetPID()                            << endl
        << "   TkID  = "  << tkid                                << endl
        << "   TkIDC = "  << tkidc                               << endl
        << "   Nclht = "  << nchits                              << endl;
   if (php) {
      cerr << "   Xpre  = (" << php->GetPosition().X() << ", "
                             << php->GetPosition().Y() << ", "
                             << php->GetPosition().Z() << ") [cm]"
           << endl;
   }

   if (fTrackPtr)   cerr << "   E(t) = " << fTrackPtr->GetEnergy()
                                         << "  [GeV]" << endl;
   if (php)         cerr << "   P(p) = " << Mph
                                         << "  [GeV]" << endl;
   if (php)         cerr << "   E(p) = " << Eph
                                         << "  [GeV]" << endl;
   if (fClusterPtr) cerr << "   E(c) = " << fClusterPtr->GetEnergy()
                                         << "  [GeV]" << endl;
   if (fBreakPointPtr) {
        cerr << "   fBreakPointPtr = " << fBreakPointPtr << endl;
        cerr << "   Bloc = " << (Block ? "Y" : "N") << endl;
        cerr << "   Ploc = " << (Plock ? "Y" : "N") << endl;
   }
   if (opt[0] == 'D') {
      if (fClusterPtr) {
         Int_t nhits = 0;
         TIter next(fClusterPtr);
         S4CALHit *chitp = 0;
         while ((chitp = static_cast<S4CALHit *>(next()))) {
            cerr << " ----- Hit No. " << nhits++ << endl;
            chitp->DebugPrint();
         }
      }
   }
}

//=====================================================================
//* Draw --------------------------------------------------------------  
                                          
void S4PFObject::Draw(Int_t color, const Char_t *opt)
{ 
   if (!gPad) return;
   TString option = opt;
   option.ToLower();

   if (fTrackPtr && !option.Contains("nt")) fTrackPtr->Draw(color, opt);

   if (!fClusterPtr || !fClusterPtr->GetEntries()) return;
   gPad->cd();

   if (option.Contains("p")) {
      S4VExactHit *ehp  = static_cast<S4CALHit *>(fClusterPtr->At(0))
                                              ->GetExactHitPtr();

      S4CALPreHit  *prep = fClusterPtr 
                         ? static_cast<S4CALExactHit *>(ehp)->GetPreHitPtr()
                         : 0;
      S4TPCPostHit *posp = prep ? prep->GetPostHitPtr() : 0;

      S4VExactHit  *php  = static_cast<S4VExactHit *>(posp) 
                         ? static_cast<S4VExactHit *>(posp)
                         : static_cast<S4VExactHit *>(prep);

      TPolyMarker3D *pm3dpp = new TPolyMarker3D(1);
      pm3dpp->SetBit(kCanDelete);
      pm3dpp->SetMarkerColor(color);
      pm3dpp->SetMarkerSize(2);
      pm3dpp->SetPoint(0, php->GetPosition().X(),
                          php->GetPosition().Y(),
                          php->GetPosition().Z());
      Int_t pdgc = php->GetPDGCode();
      switch (TMath::Abs(pdgc)) {
         case 11:
            pm3dpp->SetMarkerStyle(20); // electron (solid circle)
            break;
         case 13:
            pm3dpp->SetMarkerStyle(21); // muon (solid square)
            break;
         case 22:
            pm3dpp->SetMarkerStyle(24); // photon (open circle)
            break;
         case 211:
            pm3dpp->SetMarkerStyle(22); // pi (solid triangle)
            break;
         case 130:
            pm3dpp->SetMarkerStyle(25); // K0L (open square)
            break;
         case 2212:
            pm3dpp->SetMarkerStyle(29); // proton (solid star)
            break;
         case 2112:
            pm3dpp->SetMarkerStyle(30); // neutron (open star)
            break;
         default:
            pm3dpp->SetMarkerStyle(28); // unkonwn (open cross)
            break;
      }
      pm3dpp->Draw();
   }
   gPad->Update();

   fClusterPtr->Draw(color, opt);
}
