//*************************************************************************   
//* ---------------------                      
//* S4V0Finder Class
//* ---------------------   
//* (Description) 
//*     represent Particle Flow Object.
//* (Requires)
//*     S4CALCluster, S4HYBTrack
//* (Provides)  
//*     S4V0Finder            : made object class
//* (Update Record)
//*     2006/06/20  S.Yamamoto  Original version.
//*************************************************************************

#include "S4V0Finder.h"
#include "S4PFOMakerBuf.h"
#include "JSFVertexing.h"
#include "JSFSteer.h"
#include "JSFBasicClasses.h"
#include "JSFHelicalTrack.h"

#include <iomanip> //061118
#include <fstream>  //061118
#if __GNUC__ >= 5 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3)
#include <cstdlib>
#endif

#define __DEBUG__

ClassImp(S4V0Finder)

//=====================================================================
//* constructor -------------------------------------------------------

S4V0Finder::S4V0Finder()
{
}

//=====================================================================
//* destructor --------------------------------------------------------

S4V0Finder::~S4V0Finder()
{
}

//=====================================================================       
//* FindV0s -----------------------------------------------------------        

void S4V0Finder::FindV0s()
{
   TObjArray *pfosp = ((S4PFOMakerBuf *)gJSF->FindModule("S4PFOMaker")->EventBuf())->GetOutputBufPtr();
   if (!pfosp || !pfosp->GetEntries()) return;

   TObjArray  v0s;

   TIter next(pfosp);
   S4PFObject *pfp = 0;
   while ((pfp = (S4PFObject *)next())) {
      if (pfp->IsLocked()) continue;		// skip if locked
      S4HYBTrack   *htkp    = pfp->GetHYBTrackPtr();
      if (!htkp) continue; 			// skip if neutral
      Int_t         j4tkid  = htkp->GetTrackID();
      S4BreakPoint *bpp     = S4BreakPoint::GetBreakPointPtr(j4tkid);
      if (!bpp) continue;			// skip if orphan
      Int_t         j4tkidp = bpp->GetParentID();
      S4BreakPoint *pbpp    = S4BreakPoint::GetBreakPointPtr(j4tkidp);
      if (!pbpp) continue;			// skip if primary
      Int_t pid  = TMath::Abs(bpp ->GetParticleID());
      Int_t ppid = TMath::Abs(pbpp->GetParticleID()); 	// 310:K0s, 3122:Lambda, 22:gamma
      if (ppid == 310 || ppid == 3122 || (pid == 11 && ppid == 22)) {
        TObjArray     children;  	  	// pfo TObjArray 
        TIter         nextchild(pbpp);   	// *pbpp: TObjArray -->  iterator 
        S4BreakPoint *cbpp;    			// pointer to child BreakPoint
        while ((cbpp = static_cast<S4BreakPoint *>(nextchild()))) { 
          if (cbpp->IsLocked()) continue; 
          S4PFObject *cpfop = static_cast<S4PFObject *>(cbpp->GetPFOPtr());
          if (cpfop && !(cpfop->IsLocked())) {
            if (cpfop->GetHYBTrackPtr()) children.Add(cpfop); // add this pfo as a V0 child
          }
        }
        if (children.GetEntries() == 2) {    	// found a V0
#if 1
          v0s.Add(new S4PFObject(&children));
          pbpp->LockAllDescendants();		// lock all desendant BPs & PFOs
#else
          S4PFObject *v0pfop = new S4PFObject(&children);
	  TVector3 xv, pv;
	  Double_t e;
	  Double_t quality = FitV0(*v0pfop, xv, pv, e);
	  static const Double_t kQualityCut = 0.5;
	  if (quality < kQualityCut) {
            v0s.Add(v0pfop);
            pbpp->LockAllDescendants();		// lock all desendant BPs & PFOs
	  } else {
            delete v0pfop;
	  }
#endif
        }
      }  
   }
   if (v0s.GetEntries()) {
      TIter nextv0(&v0s);
      while ((pfp = (S4PFObject *)nextv0())) pfosp->Add(pfp);
   }
}

//=====================================================================
//* FitV0 -------------------------------------------------------------

Double_t S4V0Finder::FitV0(S4PFObject &v0pfo, TVector3 &xv, TVector3 &pv, Double_t &e)
{
  TObjArray *childrenp = v0pfo.GetDaughterPFOsPtr();
  if ((!childrenp) || childrenp->GetEntries() != 2) {    //  PFO             
    // error
    cerr << "abort" << endl;
    abort();
  }
  TObjArray &children = *childrenp;
  S4PFObject *pfo1p = static_cast<S4PFObject *>(children.At(0));
  S4PFObject *pfo2p = static_cast<S4PFObject *>(children.At(1));

  S4HYBTrack *hyb1p = pfo1p->GetHYBTrackPtr();
  S4HYBTrack *hyb2p = pfo2p->GetHYBTrackPtr();

  JSFHelicalTrack &helix1 = *hyb1p->GetHelicalTrackPtr();
  JSFHelicalTrack &helix2 = *hyb2p->GetHelicalTrackPtr();
  Double_t quality;
  xv = GetPosition(helix1, helix2, quality);
  pv = GetMomentum(helix1, helix2);
  pfo1p->SetPosition(xv);
  pfo2p->SetPosition(xv);

  // Get e
  Double_t m1 = pfo1p->GetMass();
  Double_t m2 = pfo2p->GetMass();
  e = TMath::Sqrt(m1*m1 + pfo1p->GetMag2()) + TMath::Sqrt(m2*m2 + pfo2p->GetMag2());;
#ifdef __DEBUG__
  cerr << " ------------------------------------------------------"              << endl
       << " pid  = "  << v0pfo.GetBreakPointPtr()->GetParticleID()               << endl
       << " qual = "  << quality                                                 << endl
       << " xv   = (" << xv.X() << "," << xv.Y() << "," << xv.Z() << ")"         << endl
       << " pv   = (" << pv.X() << "," << pv.Y() << "," << pv.Z() << ")"         << endl
       << " ev   = "  << e                                                       << endl
       << " mv   = "  << TMath::Sqrt(e*e - pv.Mag2())                            << endl
       << " xvbp = (" << pfo1p->GetBreakPointPtr()->GetPosition().X()     << "," 
                      << pfo1p->GetBreakPointPtr()->GetPosition().Y()     << "," 
		      << pfo1p->GetBreakPointPtr()->GetPosition().Z()     << ")" << endl
       << " pvbp = (" << v0pfo.GetBreakPointPtr()->GetFourMomentum().Px() << "," 
                      << v0pfo.GetBreakPointPtr()->GetFourMomentum().Px() << "," 
		      << v0pfo.GetBreakPointPtr()->GetFourMomentum().Px() << ")" << endl
       << " evbp = "  << v0pfo.GetBreakPointPtr()->GetFourMomentum().E()         << endl
       << " mvbp = "  << v0pfo.GetBreakPointPtr()->GetFourMomentum().Mag()       << endl
       << " m1   = "  << m1                                                      << endl
       << " m2   = "  << m2                                                      << endl
       << " pf1p = "  << pfo1p                                                   << endl
       << " pf2p = "  << pfo2p                                                   << endl
       << " ------------------------------------------------------"              << endl;
  if (quality>1.e1) {
	  cerr << " helix1:" << endl; helix1.Print();
	  cerr << " helix2:" << endl; helix2.Print();
  }
#endif

  return quality;
}

//=====================================================================
//* GetPostion --------------------------------------------------------

TVector3 S4V0Finder::GetPosition(JSFHelicalTrack &helix1, JSFHelicalTrack &helix2, Double_t &quality)
{
  JSFVertexing     vtx(2); // create vertex fitter
  static const Double_t kEpsilon = 1.e0;
  vtx.SetPairEpsilon(kEpsilon);
  vtx.SetTrack(0,helix1);   // set helx1 to vertex fitter
  vtx.SetTrack(1,helix2);   // set helx2 to vertex fitter
  quality = vtx.FindV0(); // distance of z-direction for track in vertex

  TVector3 xv;
  vtx.GetVertex(xv);

  return xv;
}

//=====================================================================
//* GetMomentum -------------------------------------------------------

TVector3 S4V0Finder::GetMomentum(JSFHelicalTrack &helix1, JSFHelicalTrack &helix2)
{
  Double_t quality;
  TVector3 xv = GetPosition(helix1,helix2,quality);
  Double_t phi1 = helix1.GetDeflectionAngle2D(JSF2DV(xv.X(),xv.Y()));
  Double_t phi2 = helix2.GetDeflectionAngle2D(JSF2DV(xv.X(),xv.Y()));

  JSF3DV jpv = helix1.GetMomentum(phi1) + helix2.GetMomentum(phi2); 
  TVector3 pv(jpv.x,jpv.y,jpv.z);

  return pv;
}
