//*************************************************************************
//* ---------------------
//* U4CDCVectorHit Class
//* ---------------------
//* (Description)
//*	Class to store "U4CDCHitWire"s for a single cell.
//* (Requires)
//*	TObjArray
//* (Provides)
//*	U4CDCVectorHit		: Array containing "U4CDCHitWire"s
//* (Update Record)
//*	2002/11/15  R.Kuboshima    Original Ver.
//*************************************************************************

#include <iostream>
#include <iomanip>
#include "TMath.h"
#include "U4CDCVectorHit.h"

ClassImp(U4CDCVectorHit)

//=====================================================================
//* constructor -------------------------------------------------------

U4CDCVectorHit::U4CDCVectorHit() 
      : TObjArray(5),
	fLayerNumber(0), fCellID(0),
	fTrackPtr(0), fDone(kFALSE), fTrackType(0), fChi2(0.),
        fa(0.), fb(0.), fEaa(0.), fEba(0.), fEbb(0.)
{
}

U4CDCVectorHit::U4CDCVectorHit(const U4CDCVectorHit &t, Bool_t mirror) 
      : TObjArray(5), TAttLockable(t),
	fLayerNumber(t.fLayerNumber), fCellID(t.fCellID),
        fTrackPtr(t.fTrackPtr), 
        fDone(t.fDone), fTrackType(t.fTrackType), fChi2(t.fChi2),
        fa(t.fa), fb(t.fb), fEaa(t.fEaa), fEba(t.fEba), fEbb(t.fEbb)
{
       	TIter next(&t);
        U4CDCTrackHit *hitp;
	while ((hitp = (U4CDCTrackHit *)next())) {
		Add(new U4CDCTrackHit(*hitp));
	}
	SetOwner();
        if (mirror) Mirror();
}


//=====================================================================
//* destructor --------------------------------------------------------

U4CDCVectorHit::~U4CDCVectorHit()
{
}

//=====================================================================
//* Compare -----------------------------------------------------------

Int_t U4CDCVectorHit::Compare(const TObject *obj) const
{
	Int_t cell  = fCellID;
	Int_t cello = ((U4CDCVectorHit *)obj)->fCellID;
   	if (cell < cello) return -1;
   	else if (cell > cello) return 1;
   	else return 0;
}

//=====================================================================
//* Add ----------------------------------------------------------
void U4CDCVectorHit::Add(TObject *obj)
{
        TObjArray::Add(obj);
        U4CDCTrackHit *hitp = (U4CDCTrackHit *)obj;
        if (!hitp->IsLocked()) fDone = kFALSE;
        if (GetEntries() == 1) {
           fLayerNumber = hitp->GetHitPtr()->GetLayerNumber();
           fCellNumber  = hitp->GetHitPtr()->GetCellNumber();
           fCellID  = hitp->GetHitPtr()->GetWirePtr()->GetCellID();
        }
	const_cast<U4CDCTrackHit *>(hitp)->SetVectorHitPtr(this);
}

//=====================================================================
//* MirrorVHit --------------------------------------------------------

void U4CDCVectorHit::Mirror()
{
	TIter next(this);
        U4CDCTrackHit *tkhitp;
        while ((tkhitp = (U4CDCTrackHit *)next())) {
        	Int_t lrsel = (tkhitp->GetLeftRight() == -1 ? 1 : -1);
                tkhitp->SetLeftRight(lrsel);
	}
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void U4CDCVectorHit::DebugPrint(const Char_t *opt)
{
	std::cerr << "CellID = " << fCellID 
		  << "  NEntries = " << GetEntries() << std::endl;
        if (strstr(opt,"Detailed")) {
           if (!fDone) Update();
           std::cerr.setf(std::ios::right);
           std::cerr.setf(std::ios::fixed);
           std::cerr << " Type = " << std::setw(2) << fTrackType
                     << " chi2 = " << std::setw(6) << std::setprecision(3) << fChi2
                     << " (a,b) = (" << std::setw(6) << std::setprecision(3) << fa
                     << "," << std::setw(6) << std::setprecision(3) << fb << ")"
                     << std::endl;
        }

#if 1
	TIter next(this);
	U4CDCTrackHit *trackhit;
	Int_t n = 0;
	while ((trackhit = (U4CDCTrackHit *)next())) {
		std::cerr << std::setw(3) << ++n << " ";
		trackhit->GetHitPtr()->DebugPrint(opt);
	}
#endif
}

//=====================================================================
//* Fit ----------------------------------------------------------

void U4CDCVectorHit::Fit()
{
        Double_t det, x, y, rs2;
        Double_t syy = 0.;
        Double_t sy  = 0.;
        Double_t s   = 0.;
        Double_t sxy = 0.;
        Double_t sx  = 0.;
        TIter next(this);
        U4CDCTrackHit *tkhitp;
        while ((tkhitp = (U4CDCTrackHit *)next())) {
                U4CDCTrackHit  &tkhit = *tkhitp;
                U4CDCHit       &hit   = *tkhitp->GetHitPtr();
                if (tkhit.IsLocked()) continue;
                rs2  = 1./TMath::Power(hit.DXdrift(),2);
                x    = hit.Xwire() + tkhit.GetLeftRight() * hit.Xdrift();
                y    = hit.Ywire();
                syy += rs2*y*y;
                sy  += rs2*y;
                s   += rs2;
                sxy += rs2*x*y;
                sx  += rs2*x;
        }
        det  = syy*s -sy*sy;
        fa   = (  s*sxy -  sy*sx)/det;
        fb   = (-sy*sxy + syy*sx)/det;
        fEaa =   s/det;
        fEba = -sy/det;
        fEbb = syy/det;

        fChi2 = 0.;
        next.Reset();
        while ((tkhitp = (U4CDCTrackHit *)next())) {
                U4CDCTrackHit  &tkhit  = *tkhitp;
                U4CDCHit       &hit    = *tkhit.GetHitPtr();
                x = hit.Xwire() + tkhit.GetLeftRight() * hit.Xdrift();
                y = hit.Ywire();
                tkhit.SetResidual(x - fa*y - fb);
                fChi2 += TMath::Power(tkhit.GetResidual()/hit.DXdrift(),2);
        }

}


