//*************************************************************************
//* ===================
//*  U4ITKalDetector Class
//* ===================
//*
//* (Description)
//* (Requires)
//* (Provides)
//* (Update Recored)
//*   2004/12/16  A.Yamaguchi       Original version.
//*
//*************************************************************************

#include <iostream>
#include <vector>
#include "JSFSteer.h"

#include "U4ITKalDetector.h"
#include "U4ITMeasLayer.h"
#include "U4ITFBMeasLayer.h"
#if __GNUC__ >= 5 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3)
#include <cstdlib>
#endif
using namespace std;

ClassImp(U4ITKalDetector)

//=====================================================================
//* constructor -------------------------------------------------------

U4ITKalDetector::U4ITKalDetector(Int_t n)
               : U4VKalDetector(n), fIsFBOn(kTRUE)
{
   SetOwner();
}

U4ITKalDetector::U4ITKalDetector(const TString &flag, Int_t n)
               : U4VKalDetector(n), fIsFBOn(kTRUE)
{
   if (flag != "FillLayer") {
      std::cerr << "<<< ERROR!! >>> flag should be \"FillLayer\"" << std::endl;
      abort();
   }
   Initialize();
   SetOwner();
}

//=====================================================================
//* destructor --------------------------------------------------------

U4ITKalDetector::~U4ITKalDetector()
{
   Clear();
}

//=====================================================================
//* Initialize --------------------------------------------------------

void U4ITKalDetector::Initialize()
{
   Double_t a, z, dens, radl;
   TMaterial &air = *new TMaterial("ITAir", "",
                                   a    = 14.00674*0.7+15.9994*0.3,
                                   z    = 7.3,
                                   dens = 1.205e-3,
                                   radl = 3.42e4, 0.);
   TMaterial &si  = *new TMaterial("ITSi", "",
                                   a = 28.0855,
                                   z = 14.,
                                   dens = 2.33,
                                   radl = 9.36, 0.);

   JSFEnv &env = *gJSF->Env();

   // Barrel part ===============================================
   Int_t            nlayers = env.GetValue("J4IT.NBarrel", 4);
   vector<Double_t> rin     = env.GetDValue("J4IT.Barrel.InnerRadius","9.0 16.0 23.0 30.0", nlayers);
   vector<Double_t> hzlen   = env.GetDValue("J4IT.Barrel.HalfZ","18.5 33.0 47.5 61.0", nlayers);
   Double_t         thick   = env.GetValue("J4IT.Layer.Thickness", 0.056160);

   Bool_t active = kTRUE;
   Bool_t dummy  = kFALSE;
   // create measurement layers of inner tracker
   for (Int_t layer = 0; layer < nlayers; layer++) {
      Add(new U4ITMeasLayer(air, si, rin[layer], hzlen[layer], active));
      Add(new U4ITMeasLayer(si, air, rin[layer] + thick, hzlen[layer], dummy));
   }

   if (!IsFBOn()) return;
   // Fwd/Bwd part ===============================================
   Int_t            n_fb  = env.GetValue("J4IT.NEndcap", 7);
   vector<Double_t> ri_fb = env.GetDValue("J4IT.Endcap.Rmin", "3. 5.5 8. 8. 8. 8. 8.", n_fb);
   vector<Double_t> ro_fb = env.GetDValue("J4IT.Endcap.Rmax", "14. 17. 25. 32. 32. 32. 32.", n_fb);
   vector<Double_t> z_fb  = env.GetDValue("J4IT.Endcap.Zpos", "20. 35. 50. 65. 90. 115. 140.", n_fb);
   Int_t            n_dm  = static_cast<Int_t>((env.GetValue("J4.TPC.HalfZ", 260.) - z_fb[n_fb - 1])
                          / (z_fb[n_fb - 1] / n_fb));
   Double_t         dz_dm = (env.GetValue("J4.TPC.HalfZ", 260.) - z_fb[n_fb - 1])/(n_dm + 1);

   vector<Double_t> ro_al(n_fb + n_dm);
   // create measurement layers of inner tracker
   for (Int_t layer = 0; layer < n_fb + n_dm; layer++) {
      Double_t z = layer < n_fb ? z_fb[layer] : z_fb[n_fb - 1] + dz_dm*(layer - n_fb + 1);
      TVector3 xc1(0., 0., z);
      TVector3 xc2(0., 0., z + thick);
      TVector3 xc3(-xc1);
      TVector3 xc4(-xc2);
      if (layer < n_fb) {
         // layer should be bigger than its inner layer
         // because Rout is used as a sorting policy.
         ro_al[layer] = (layer && ro_fb[layer] == ro_fb[layer - 1])
                      ? ro_al[layer - 1] + 1.e-6 : ro_fb[layer];

         Add(new U4ITFBMeasLayer(air, si, xc1, ri_fb[layer], ro_al[layer], active));
         Add(new U4ITFBMeasLayer(si, air, xc2, ri_fb[layer], ro_al[layer], dummy));
         Add(new U4ITFBMeasLayer(air, si, xc3, ri_fb[layer], ro_al[layer], active));
         Add(new U4ITFBMeasLayer(si, air, xc4, ri_fb[layer], ro_al[layer], dummy));
      } else {
         ro_al[layer] = ro_al[n_fb - 1] + 1.e-6;
         Add(new U4ITFBMeasLayer(air, air, xc1, ri_fb[n_fb - 1], ro_al[layer], dummy));
         Add(new U4ITFBMeasLayer(air, air, xc3, ri_fb[n_fb - 1], ro_al[layer], dummy));
      }
   }
}

//=====================================================================
//* GetMeasLayerPtr ---------------------------------------------------

const U4VMeasLayer *U4ITKalDetector::GetMeasLayerPtr(Int_t index) const
{
   return dynamic_cast<U4VMeasLayer *>(At(index * 2));
}

